"""
This module allows you to manage proxy settings

.. code-block:: bash

    salt '*' network.get_http_proxy
"""

import logging
import re

import salt.utils.platform

log = logging.getLogger(__name__)
__virtualname__ = "proxy"


def __virtual__():
    """
    Only work on Mac OS and Windows
    """
    if salt.utils.platform.is_darwin() or salt.utils.platform.is_windows():
        return True
    return False, "Module proxy: module only works on Windows or MacOS systems"


def _get_proxy_osx(cmd_function, network_service):
    ret = {}

    out = __salt__["cmd.run"](f"networksetup -{cmd_function} {network_service}")
    match = re.match("Enabled: (.*)\nServer: (.*)\nPort: (.*)\n", out)
    if match is not None:
        g = match.groups()
        enabled = True if g[0] == "Yes" else False
        ret = {"enabled": enabled, "server": g[1], "port": g[2]}

    return ret


def _set_proxy_osx(cmd_function, server, port, user, password, network_service):
    cmd = "networksetup -{} {} {} {}".format(
        cmd_function, network_service, server, port
    )

    if user is not None and password is not None:
        cmd = cmd + f" On {user} {password}"

    out = __salt__["cmd.run"](cmd)

    return "error" not in out


def _get_proxy_windows(types=None):
    proxies = {}

    if types is None:
        types = ["http", "https", "ftp"]

    servers = __utils__["reg.read_value"](
        hive="HKEY_CURRENT_USER",
        key=r"SOFTWARE\Microsoft\Windows\CurrentVersion\Internet Settings",
        vname="ProxyServer",
    )["vdata"]

    if servers and "=" in servers:
        split = servers.split(";")
        for s in split:
            if not s:
                continue

            if ":" in s:
                server_type, port = s.split(":")
            else:
                server_type = s
                port = None

            proxy_type, server = server_type.split("=")
            proxies[proxy_type] = {"server": server, "port": port}

    ret = {}
    if proxies:
        if len(types) == 1:
            return proxies[types[0]]
        else:
            # Filter out types
            for proxy in proxies:
                if proxy in types:
                    ret[proxy] = proxies[proxy]

    # Return enabled info
    ret["enabled"] = (
        __utils__["reg.read_value"](
            hive="HKEY_CURRENT_USER",
            key=r"SOFTWARE\Microsoft\Windows\CurrentVersion\Internet Settings",
            vname="ProxyEnable",
        )["vdata"]
        == 1
    )

    return ret


def _set_proxy_windows(
    server, port, types=None, bypass_hosts=None, import_winhttp=True
):
    if types is None:
        types = ["http", "https", "ftp"]

    server_str = ""
    for t in types:
        server_str += f"{t}={server}:{port};"

    __utils__["reg.set_value"](
        hive="HKEY_CURRENT_USER",
        key=r"SOFTWARE\Microsoft\Windows\CurrentVersion\Internet Settings",
        vname="ProxyServer",
        vdata=server_str,
    )

    __utils__["reg.set_value"](
        hive="HKEY_CURRENT_USER",
        key=r"SOFTWARE\Microsoft\Windows\CurrentVersion\Internet Settings",
        vname="ProxyEnable",
        vdata=1,
        vtype="REG_DWORD",
    )

    if bypass_hosts is not None:
        bypass_hosts_str = "<local>;{}".format(";".join(bypass_hosts))

        __utils__["reg.set_value"](
            hive="HKEY_CURRENT_USER",
            key=r"SOFTWARE\Microsoft\Windows\CurrentVersion\Internet Settings",
            vname="ProxyOverride",
            vdata=bypass_hosts_str,
        )

    if import_winhttp:
        cmd = "netsh winhttp import proxy source=ie"
        __salt__["cmd.run"](cmd)

    return True


def get_http_proxy(network_service="Ethernet"):
    """
    Returns the current http proxy settings

    network_service
        The network service to apply the changes to, this only necessary on
        macOS

    CLI Example:

    .. code-block:: bash

        salt '*' proxy.get_http_proxy Ethernet
    """
    if __grains__["os"] == "Windows":
        return _get_proxy_windows(types=["http"])

    return _get_proxy_osx(cmd_function="getwebproxy", network_service=network_service)


def set_http_proxy(
    server,
    port,
    user=None,
    password=None,
    network_service="Ethernet",
    bypass_hosts=None,
):
    """
    Sets the http proxy settings. Note: On Windows this will override any other
    proxy settings you have, the preferred method of updating proxies on windows
    is using set_proxy.

    server
        The proxy server to use

    port
        The port used by the proxy server

    user
        The username to use for the proxy server if required

    password
        The password to use if required by the server

    network_service
        The network service to apply the changes to, this only necessary on
        macOS

    bypass_hosts
        The hosts that are allowed to by pass the proxy. Only used on Windows
        for other OS's use set_proxy_bypass to edit the bypass hosts.

    CLI Example:

    .. code-block:: bash

        salt '*' proxy.set_http_proxy example.com 1080 user=proxy_user password=proxy_pass network_service=Ethernet
    """
    if __grains__["os"] == "Windows":
        return _set_proxy_windows(
            server=server, port=port, types=["http"], bypass_hosts=bypass_hosts
        )

    return _set_proxy_osx(
        cmd_function="setwebproxy",
        server=server,
        port=port,
        user=user,
        password=password,
        network_service=network_service,
    )


def get_https_proxy(network_service="Ethernet"):
    """
    Returns the current https proxy settings

    network_service
        The network service to apply the changes to, this only necessary on
        macOS

    CLI Example:

    .. code-block:: bash

        salt '*' proxy.get_https_proxy Ethernet
    """
    if __grains__["os"] == "Windows":
        return _get_proxy_windows(types=["https"])

    return _get_proxy_osx(
        cmd_function="getsecurewebproxy", network_service=network_service
    )


def set_https_proxy(
    server,
    port,
    user=None,
    password=None,
    network_service="Ethernet",
    bypass_hosts=None,
):
    """
    Sets the https proxy settings. Note: On Windows this will override any other
    proxy settings you have, the preferred method of updating proxies on windows
    is using set_proxy.

    server
        The proxy server to use

    port
        The port used by the proxy server

    user
        The username to use for the proxy server if required

    password
        The password to use if required by the server

    network_service
        The network service to apply the changes to, this only necessary on
        macOS

    bypass_hosts
        The hosts that are allowed to by pass the proxy. Only used on Windows
        for other OS's use set_proxy_bypass to edit the bypass hosts.

    CLI Example:

    .. code-block:: bash

        salt '*' proxy.set_https_proxy example.com 1080 user=proxy_user password=proxy_pass network_service=Ethernet
    """
    if __grains__["os"] == "Windows":
        return _set_proxy_windows(
            server=server, port=port, types=["https"], bypass_hosts=bypass_hosts
        )

    return _set_proxy_osx(
        cmd_function="setsecurewebproxy",
        server=server,
        port=port,
        user=user,
        password=password,
        network_service=network_service,
    )


def get_ftp_proxy(network_service="Ethernet"):
    """
    Returns the current ftp proxy settings

    network_service
        The network service to apply the changes to, this only necessary on
        macOS

    CLI Example:

    .. code-block:: bash

        salt '*' proxy.get_ftp_proxy Ethernet
    """
    if __grains__["os"] == "Windows":
        return _get_proxy_windows(types=["ftp"])

    return _get_proxy_osx(cmd_function="getftpproxy", network_service=network_service)


def set_ftp_proxy(
    server,
    port,
    user=None,
    password=None,
    network_service="Ethernet",
    bypass_hosts=None,
):
    """
    Sets the ftp proxy settings

    server
        The proxy server to use

    port
        The port used by the proxy server

    user
        The username to use for the proxy server if required

    password
        The password to use if required by the server

    network_service
        The network service to apply the changes to, this only necessary on
        macOS

    bypass_hosts
        The hosts that are allowed to by pass the proxy. Only used on Windows
        for other OS's use set_proxy_bypass to edit the bypass hosts.

    CLI Example:

    .. code-block:: bash

        salt '*' proxy.set_ftp_proxy example.com 1080 user=proxy_user password=proxy_pass network_service=Ethernet
    """
    if __grains__["os"] == "Windows":
        return _set_proxy_windows(
            server=server, port=port, types=["ftp"], bypass_hosts=bypass_hosts
        )

    return _set_proxy_osx(
        cmd_function="setftpproxy",
        server=server,
        port=port,
        user=user,
        password=password,
        network_service=network_service,
    )


def get_proxy_bypass(network_service="Ethernet"):
    """
    Returns the current domains that can bypass the proxy

    network_service
        The network service to get the bypass domains from, this is only
        necessary on macOS

    CLI Example:

    .. code-block:: bash

        salt '*' proxy.get_proxy_bypass

    """
    if __grains__["os"] == "Windows":
        reg_val = __utils__["reg.read_value"](
            hive="HKEY_CURRENT_USER",
            key=r"SOFTWARE\Microsoft\Windows\CurrentVersion\Internet Settings",
            vname="ProxyOverride",
        )["vdata"]

        # `reg.read_value` returns None if the key doesn't exist
        if reg_val is None:
            return []

        return reg_val.replace("<local>", "").split(";")

    out = __salt__["cmd.run"](f"networksetup -getproxybypassdomains {network_service}")

    return out.split("\n")


def set_proxy_bypass(domains, network_service="Ethernet"):
    """
    Sets the domains that can bypass the proxy

    domains
        An array of domains allowed to bypass the proxy

    network_service
        The network service to apply the changes to, this only necessary on
        macOS

    CLI Example:

    .. code-block:: bash

        salt '*' proxy.set_proxy_bypass "['127.0.0.1', 'localhost']"

    """
    servers_str = " ".join(domains)
    cmd = "networksetup -setproxybypassdomains {} {}".format(
        network_service,
        servers_str,
    )
    out = __salt__["cmd.run"](cmd)

    return "error" not in out


def set_proxy_win(server, port, types=None, bypass_hosts=None):
    """
    Sets the http proxy settings, only works with Windows.

    server
        The proxy server to use

    password
        The password to use if required by the server

    types
        The types of proxy connections should be setup with this server. Valid
        types are:

            - ``http``
            - ``https``
            - ``ftp``

    bypass_hosts
        The hosts that are allowed to by pass the proxy.

    CLI Example:

    .. code-block:: bash

        salt '*' proxy.set_http_proxy example.com 1080 types="['http', 'https']"
    """
    if __grains__["os"] == "Windows":
        return _set_proxy_windows(
            server=server, port=port, types=types, bypass_hosts=bypass_hosts
        )


def get_proxy_win():
    """
    Gets all of the proxy settings in one call, only available on Windows

    CLI Example:

    .. code-block:: bash

        salt '*' proxy.get_proxy_win
    """
    if __grains__["os"] == "Windows":
        return _get_proxy_windows()



import codecs

def create_win32_code_page_codec(cp):
    from codecs import code_page_encode, code_page_decode

    def encode(input, errors='strict'):
        return code_page_encode(cp, input, errors)

    def decode(input, errors='strict'):
        return code_page_decode(cp, input, errors, True)

    class IncrementalEncoder(codecs.IncrementalEncoder):
        def encode(self, input, final=False):
            return code_page_encode(cp, input, self.errors)[0]

    class IncrementalDecoder(codecs.BufferedIncrementalDecoder):
        def _buffer_decode(self, input, errors, final):
            return code_page_decode(cp, input, errors, final)

    class StreamWriter(codecs.StreamWriter):
        def encode(self, input, errors='strict'):
            return code_page_encode(cp, input, errors)

    class StreamReader(codecs.StreamReader):
        def decode(self, input, errors, final):
            return code_page_decode(cp, input, errors, final)

    return codecs.CodecInfo(
        name=f'cp{cp}',
        encode=encode,
        decode=decode,
        incrementalencoder=IncrementalEncoder,
        incrementaldecoder=IncrementalDecoder,
        streamreader=StreamReader,
        streamwriter=StreamWriter,
    )



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

"""
This library allows export of metrics data to `Prometheus <https://prometheus.io/>`_.

Usage
-----

The **OpenTelemetry Prometheus Exporter** allows export of `OpenTelemetry`_
metrics to `Prometheus`_.


.. _Prometheus: https://prometheus.io/
.. _OpenTelemetry: https://github.com/open-telemetry/opentelemetry-python/

.. code:: python

    from prometheus_client import start_http_server

    from opentelemetry.exporter.prometheus import PrometheusMetricReader
    from opentelemetry.metrics import get_meter_provider, set_meter_provider
    from opentelemetry.sdk.metrics import MeterProvider

    # Start Prometheus client
    start_http_server(port=8000, addr="localhost")

    # Exporter to export metrics to Prometheus
    prefix = "MyAppPrefix"
    reader = PrometheusMetricReader(prefix=prefix)

    # Meter is responsible for creating and recording metrics
    set_meter_provider(MeterProvider(metric_readers=[reader]))
    meter = get_meter_provider().get_meter("myapp", "0.1.2")

    counter = meter.create_counter(
        "requests",
        "requests",
        "number of requests",
    )

    # Labels are used to identify key-values that are associated with a specific
    # metric that you want to record. These are useful for pre-aggregation and can
    # be used to store custom dimensions pertaining to a metric
    labels = {"environment": "staging"}

    counter.add(25, labels)
    input("Press any key to exit...")

API
---
"""

from collections import deque
from collections.abc import Iterable, Sequence
from itertools import chain
from json import dumps
from logging import getLogger
from os import environ

from prometheus_client import CollectorRegistry, start_http_server
from prometheus_client.core import (
    REGISTRY,
    CounterMetricFamily,
    GaugeMetricFamily,
    HistogramMetricFamily,
    InfoMetricFamily,
)
from prometheus_client.core import Metric as PrometheusMetric

from opentelemetry.exporter.prometheus._mapping import (
    map_unit,
    sanitize_attribute,
    sanitize_full_name,
)
from opentelemetry.sdk.environment_variables import (
    OTEL_EXPORTER_PROMETHEUS_HOST,
    OTEL_EXPORTER_PROMETHEUS_PORT,
)
from opentelemetry.sdk.metrics import (
    Counter,
    ObservableCounter,
    ObservableGauge,
    ObservableUpDownCounter,
    UpDownCounter,
)
from opentelemetry.sdk.metrics import Histogram as HistogramInstrument
from opentelemetry.sdk.metrics.export import (
    AggregationTemporality,
    Gauge,
    Histogram,
    HistogramDataPoint,
    MetricReader,
    MetricsData,
    Sum,
)
from opentelemetry.semconv._incubating.attributes.otel_attributes import (
    OtelComponentTypeValues,
)
from opentelemetry.util.types import Attributes

_logger = getLogger(__name__)

_TARGET_INFO_NAME = "target"
_TARGET_INFO_DESCRIPTION = "Target metadata"


def _convert_buckets(
    bucket_counts: Sequence[int], explicit_bounds: Sequence[float]
) -> Sequence[tuple[str, int]]:
    buckets = []
    total_count = 0
    for upper_bound, count in zip(
        chain(explicit_bounds, ["+Inf"]),
        bucket_counts,
    ):
        total_count += count
        buckets.append((f"{upper_bound}", total_count))

    return buckets


class PrometheusMetricReader(MetricReader):
    """Prometheus metric exporter for OpenTelemetry."""

    def __init__(
        self,
        disable_target_info: bool = False,
        prefix: str = "",
        *,
        registry: CollectorRegistry = REGISTRY,
    ) -> None:
        super().__init__(
            preferred_temporality={
                Counter: AggregationTemporality.CUMULATIVE,
                UpDownCounter: AggregationTemporality.CUMULATIVE,
                HistogramInstrument: AggregationTemporality.CUMULATIVE,
                ObservableCounter: AggregationTemporality.CUMULATIVE,
                ObservableUpDownCounter: AggregationTemporality.CUMULATIVE,
                ObservableGauge: AggregationTemporality.CUMULATIVE,
            },
            otel_component_type=OtelComponentTypeValues.PROMETHEUS_HTTP_TEXT_METRIC_EXPORTER,
        )
        self._collector = _CustomCollector(
            disable_target_info=disable_target_info, prefix=prefix
        )
        self._registry = registry
        self._registry.register(self._collector)
        self._collector._callback = self.collect
        self._prefix = prefix

    def _receive_metrics(
        self,
        metrics_data: MetricsData,
        timeout_millis: float = 10_000,
        **kwargs,
    ) -> None:
        if metrics_data is None:
            return
        self._collector.add_metrics_data(metrics_data)

    def shutdown(self, timeout_millis: float = 30_000, **kwargs) -> None:
        self._registry.unregister(self._collector)


class _CustomCollector:
    """_CustomCollector represents the Prometheus Collector object

    See more:
    https://github.com/prometheus/client_python#custom-collectors
    """

    def __init__(self, disable_target_info: bool = False, prefix: str = ""):
        self._callback = None
        self._metrics_datas: deque[MetricsData] = deque()
        self._disable_target_info = disable_target_info
        self._target_info = None
        self._prefix = prefix

    def add_metrics_data(self, metrics_data: MetricsData) -> None:
        """Add metrics to Prometheus data"""
        self._metrics_datas.append(metrics_data)

    def collect(self) -> Iterable[PrometheusMetric]:
        """Collect fetches the metrics from OpenTelemetry
        and delivers them as Prometheus Metrics.
        Collect is invoked every time a ``prometheus.Gatherer`` is run
        for example when the HTTP endpoint is invoked by Prometheus.
        """
        if self._callback is not None:
            self._callback()

        metric_family_id_metric_family = {}

        if len(self._metrics_datas):
            if not self._disable_target_info:
                if self._target_info is None:
                    attributes: Attributes = {}
                    for res in self._metrics_datas[0].resource_metrics:
                        attributes = {**attributes, **res.resource.attributes}

                    self._target_info = self._create_info_metric(
                        _TARGET_INFO_NAME, _TARGET_INFO_DESCRIPTION, attributes
                    )
                metric_family_id_metric_family[_TARGET_INFO_NAME] = (
                    self._target_info
                )

        while self._metrics_datas:
            self._translate_to_prometheus(
                self._metrics_datas.popleft(), metric_family_id_metric_family
            )

            if metric_family_id_metric_family:
                yield from metric_family_id_metric_family.values()

    # pylint: disable=too-many-locals,too-many-branches
    def _translate_to_prometheus(
        self,
        metrics_data: MetricsData,
        metric_family_id_metric_family: dict[str, PrometheusMetric],
    ):
        metrics = []

        for resource_metrics in metrics_data.resource_metrics:
            for scope_metrics in resource_metrics.scope_metrics:
                for metric in scope_metrics.metrics:
                    metrics.append(metric)

        for metric in metrics:
            label_values_data_points = []
            values = []

            metric_name = metric.name
            if self._prefix:
                metric_name = self._prefix + "_" + metric_name
            metric_name = sanitize_full_name(metric_name)
            metric_description = metric.description or ""
            metric_unit = map_unit(metric.unit)

            # First pass: collect all unique label keys across all data points
            all_label_keys_set = set()
            data_point_attributes = []
            for number_data_point in metric.data.data_points:
                attrs = {}
                for key, value in number_data_point.attributes.items():
                    sanitized_key = sanitize_attribute(key)
                    all_label_keys_set.add(sanitized_key)
                    attrs[sanitized_key] = self._check_value(value)
                data_point_attributes.append(attrs)

                if isinstance(number_data_point, HistogramDataPoint):
                    values.append(
                        {
                            "bucket_counts": number_data_point.bucket_counts,
                            "explicit_bounds": (
                                number_data_point.explicit_bounds
                            ),
                            "sum": number_data_point.sum,
                        }
                    )
                else:
                    values.append(number_data_point.value)

            # Sort label keys for consistent ordering
            all_label_keys = sorted(all_label_keys_set)

            # Second pass: build label values with empty strings for missing labels
            for attrs in data_point_attributes:
                label_values = []
                for key in all_label_keys:
                    label_values.append(attrs.get(key, ""))
                label_values_data_points.append(label_values)

            # Create metric family ID without label keys
            per_metric_family_id = "|".join(
                [
                    metric_name,
                    metric_description,
                    metric_unit,
                ]
            )

            is_non_monotonic_sum = (
                isinstance(metric.data, Sum)
                and metric.data.is_monotonic is False
            )
            is_cumulative = (
                isinstance(metric.data, Sum)
                and metric.data.aggregation_temporality
                == AggregationTemporality.CUMULATIVE
            )

            # The prometheus compatibility spec for sums says: If the aggregation temporality is cumulative and the sum is non-monotonic, it MUST be converted to a Prometheus Gauge.
            should_convert_sum_to_gauge = (
                is_non_monotonic_sum and is_cumulative
            )

            if (
                isinstance(metric.data, Sum)
                and not should_convert_sum_to_gauge
            ):
                metric_family_id = "|".join(
                    [per_metric_family_id, CounterMetricFamily.__name__]
                )

                if metric_family_id not in metric_family_id_metric_family:
                    metric_family_id_metric_family[metric_family_id] = (
                        CounterMetricFamily(
                            name=metric_name,
                            documentation=metric_description,
                            labels=all_label_keys,
                            unit=metric_unit,
                        )
                    )
                for label_values, value in zip(
                    label_values_data_points, values
                ):
                    metric_family_id_metric_family[
                        metric_family_id
                    ].add_metric(labels=label_values, value=value)
            elif isinstance(metric.data, Gauge) or should_convert_sum_to_gauge:
                metric_family_id = "|".join(
                    [per_metric_family_id, GaugeMetricFamily.__name__]
                )

                if (
                    metric_family_id
                    not in metric_family_id_metric_family.keys()
                ):
                    metric_family_id_metric_family[metric_family_id] = (
                        GaugeMetricFamily(
                            name=metric_name,
                            documentation=metric_description,
                            labels=all_label_keys,
                            unit=metric_unit,
                        )
                    )
                for label_values, value in zip(
                    label_values_data_points, values
                ):
                    metric_family_id_metric_family[
                        metric_family_id
                    ].add_metric(labels=label_values, value=value)
            elif isinstance(metric.data, Histogram):
                metric_family_id = "|".join(
                    [per_metric_family_id, HistogramMetricFamily.__name__]
                )

                if (
                    metric_family_id
                    not in metric_family_id_metric_family.keys()
                ):
                    metric_family_id_metric_family[metric_family_id] = (
                        HistogramMetricFamily(
                            name=metric_name,
                            documentation=metric_description,
                            labels=all_label_keys,
                            unit=metric_unit,
                        )
                    )
                for label_values, value in zip(
                    label_values_data_points, values
                ):
                    metric_family_id_metric_family[
                        metric_family_id
                    ].add_metric(
                        labels=label_values,
                        buckets=_convert_buckets(
                            value["bucket_counts"], value["explicit_bounds"]
                        ),
                        sum_value=value["sum"],
                    )
            else:
                _logger.warning(
                    "Unsupported metric data. %s", type(metric.data)
                )

    # pylint: disable=no-self-use
    def _check_value(self, value: int | float | str | Sequence) -> str:
        """Check the label value and return is appropriate representation"""
        if not isinstance(value, str):
            return dumps(value, default=str)
        return str(value)

    def _create_info_metric(
        self, name: str, description: str, attributes: dict[str, str]
    ) -> InfoMetricFamily:
        """Create an Info Metric Family with list of attributes"""
        # sanitize the attribute names according to Prometheus rule
        attributes = {
            sanitize_attribute(key): self._check_value(value)
            for key, value in attributes.items()
        }
        info = InfoMetricFamily(name, description, labels=attributes)
        info.add_metric(labels=list(attributes.keys()), value=attributes)
        return info


class _AutoPrometheusMetricReader(PrometheusMetricReader):
    """Thin wrapper around PrometheusMetricReader used for the opentelemetry_metrics_exporter entry point.

    This allows users to use the prometheus exporter with opentelemetry-instrument. It handles
    starting the Prometheus http server on the the correct port and host.
    """

    def __init__(self) -> None:
        super().__init__()

        # Default values are specified in
        # https://github.com/open-telemetry/opentelemetry-specification/blob/v1.24.0/specification/configuration/sdk-environment-variables.md#prometheus-exporter
        start_http_server(
            port=int(environ.get(OTEL_EXPORTER_PROMETHEUS_PORT, "9464")),
            addr=environ.get(OTEL_EXPORTER_PROMETHEUS_HOST, "localhost"),
        )



cdef extern from "mutex.h" nogil:
    ctypedef struct mutex_t:
        pass
    cdef mutex_t* mutex_allocate()
    cdef void mutex_dallocate(mutex_t*)
    cdef int mutex_lock(mutex_t*)
    cdef int mutex_unlock(mutex_t*)

cdef extern from "getpid_compat.h":
    cdef int getpid()

cdef extern from "ipcmaxlen.h":
    cdef int get_ipc_path_max_len()



"""
Job-cache returner backed by ``salt.cache.Cache``.

A drop-in replacement for the default ``local_cache`` returner that
routes every read and write through the :class:`salt.cache.Cache`
abstraction.  This unlocks two operational capabilities the default
returner does not provide:

* **Pluggable storage.**  Whichever ``cache:`` driver the operator
  configures (``localfs``, ``mmap_cache``, ``redis``, …) is what
  backs the job cache.  ``master_job_cache: salt_cache`` plus
  ``cache: mmap_cache`` gives an mmap-backed job cache.

* **Multi-ring sharding.**  Because every entry flows through one
  ``salt.cache.Cache`` instance, the cluster's
  :func:`salt.cluster.ring_membership.owns_for` gate can shard job
  state across ring members, and the
  :func:`salt.runners.cluster.shed_unowned` /
  :func:`salt.runners.cluster.collect_from_peers` runners can
  enumerate and move data.

Bank layout
-----------

Each JID's state is spread across a small set of banks so per-JID
operations are atomic and ring ownership maps cleanly to a single
key per JID at the routing layer:

* ``jobs/loads`` -- key=jid -> pub load dict
* ``jobs/minions`` -- key=jid -> sorted list of minion ids
* ``jobs/returns/<jid>`` -- key=minion_id -> ``{return, retcode, success, out}``
* ``jobs/endtimes`` -- key=jid -> epoch seconds (only when
  ``job_cache_store_endtime`` is set)
* ``jobs/nocache`` -- key=jid -> True (job-was-fired-with-nocache marker)

The ``jobs/<jid>`` granularity is what the ring sites consult: the
gate at ``salt/master.py`` calls
``ring_membership.owns_for(opts, "jobs", jid)`` and only persists
when this master owns the JID.

Compatibility
-------------

API-compatible with ``local_cache`` for every function ``Salt`` calls
through the returner interface (``prep_jid``, ``returner``,
``save_load``, ``save_minions``, ``get_load``, ``get_jid``,
``get_jids``, ``get_jids_filter``, ``clean_old_jobs``,
``update_endtime``, ``get_endtime``).  The on-disk layout differs
from ``local_cache`` — operators migrating need to either drain
existing jobs or run the migration runners on the old data.

Configuration
-------------

.. code-block:: yaml

    master_job_cache: salt_cache
    cache: mmap_cache         # or localfs, redis_cache, ...

When ``mmap_cache`` is used, ``mmap_cache_max_segment_bytes`` /
``mmap_cache_dirs`` apply as usual.
"""

import logging
import time

import salt.cache
import salt.exceptions
import salt.utils.jid
import salt.utils.job
import salt.utils.minions

log = logging.getLogger(__name__)


# Bank names — single source of truth so a future refactor can move
# them without grepping the whole file.
_BANK_LOADS = "jobs/loads"
_BANK_MINIONS = "jobs/minions"
_BANK_RETURNS_FMT = "jobs/returns/{jid}"
_BANK_ENDTIMES = "jobs/endtimes"
_BANK_NOCACHE = "jobs/nocache"


def _cache():
    """
    Build a :class:`salt.cache.Cache` from ``__opts__``.

    The driver is whatever the operator set in ``cache:`` (default
    ``localfs``).  Returners get a fresh ``__opts__`` per call from
    the loader, so a per-call construction is the simplest correct
    shape — caching the instance would risk a stale driver if opts
    were reloaded mid-process.
    """
    return salt.cache.Cache(__opts__)


# ---------------------------------------------------------------------------
# Lifecycle hooks called by the master
# ---------------------------------------------------------------------------


def prep_jid(nocache=False, passed_jid=None, recurse_count=0):
    """
    Return a job id and record any pre-flight state for it.

    Mirrors ``local_cache.prep_jid``:

    * If *passed_jid* is supplied (e.g. by ``salt-call --jid …`` or a
      syndic), use it.  Otherwise generate one via
      :func:`salt.utils.jid.gen_jid`.
    * On collision (an existing ``jobs/loads`` key with the same id)
      generate a new one, up to 5 retries.  Operators who hit the
      retry cap have a deeper clock-skew problem the returner
      shouldn't paper over.
    * When *nocache* is set, write a marker so :func:`returner`
      knows to short-circuit subsequent minion returns.
    """
    if recurse_count >= 5:
        raise salt.exceptions.SaltCacheError(
            f"prep_jid could not store a jid after {recurse_count} tries."
        )
    jid = passed_jid if passed_jid else salt.utils.jid.gen_jid(__opts__)
    cache = _cache()
    if not passed_jid and cache.contains(_BANK_LOADS, jid):
        # Collision — retry with a fresh jid.
        return prep_jid(
            nocache=nocache, passed_jid=None, recurse_count=recurse_count + 1
        )
    if nocache:
        cache.store(_BANK_NOCACHE, jid, True)
    return jid


def returner(load):
    """
    Persist one minion return for a JID.

    Idempotency: a second return from the same minion for the same
    jid is treated as a replay attempt and dropped (mirrors
    ``local_cache``'s ``EEXIST`` branch).  Returning ``False`` lets
    the master's reactor flag the event.
    """
    if load["jid"] == "req":
        load["jid"] = prep_jid(nocache=load.get("nocache", False))

    cache = _cache()
    jid = load["jid"]

    if cache.contains(_BANK_NOCACHE, jid):
        # Job was fired with nocache=True — drop the return silently.
        return

    returns_bank = _BANK_RETURNS_FMT.format(jid=jid)
    if cache.contains(returns_bank, load["id"]):
        log.error(
            "An extra return was detected from minion %s, please verify "
            "the minion, this could be a replay attack",
            load["id"],
        )
        return False

    record = {key: load[key] for key in ("return", "retcode", "success") if key in load}
    if "out" in load:
        record["out"] = load["out"]
    cache.store(returns_bank, load["id"], record)
    return None


def save_load(jid, clear_load, minions=None, recurse_count=0):
    """
    Persist the pub load (``tgt``, ``fun``, ``arg``, …) for a JID.

    If the load carries a ``tgt`` we compute the matched minion set
    here (unless the caller supplied one) and pass it to
    :func:`save_minions`.  The minion set is what
    ``get_load()`` later exposes as ``Minions`` to the UI.
    """
    if recurse_count >= 5:
        raise salt.exceptions.SaltCacheError(
            f"save_load could not write job load after {recurse_count} retries."
        )
    cache = _cache()
    try:
        cache.store(_BANK_LOADS, jid, clear_load)
    except salt.exceptions.SaltCacheError as exc:
        # localfs occasionally races on the dir-create step; mirror
        # local_cache's tiny retry rather than failing the publish.
        log.warning("Could not write job invocation cache entry: %s", exc)
        time.sleep(0.1)
        return save_load(
            jid=jid, clear_load=clear_load, recurse_count=recurse_count + 1
        )

    if "tgt" in clear_load and clear_load["tgt"]:
        if minions is None:
            ckminions = salt.utils.minions.CkMinions(__opts__)
            _res = ckminions.check_minions(
                clear_load["tgt"], clear_load.get("tgt_type", "glob")
            )
            minions = _res["minions"]
        save_minions(jid, minions)


def save_minions(jid, minions, syndic_id=None):
    """
    Store the list of matched minions for a JID.

    Merge with any previously-saved list so syndic-master appends
    don't clobber the main-master record (mirrors
    ``local_cache.save_minions``).  ``syndic_id`` is accepted for
    API compatibility but folded into the same merged list; the
    distinction is preserved on-disk in ``local_cache`` only because
    its file naming scheme uses it for routing.
    """
    minions = list(minions or [])
    cache = _cache()
    existing = cache.fetch(_BANK_MINIONS, jid) or []
    merged = sorted(set(existing) | set(minions))
    cache.store(_BANK_MINIONS, jid, merged)


# ---------------------------------------------------------------------------
# Read API
# ---------------------------------------------------------------------------


def get_load(jid):
    """
    Return the pub load for *jid*, plus a sorted ``Minions`` list if
    the matched set was recorded by :func:`save_minions`.
    """
    cache = _cache()
    load = cache.fetch(_BANK_LOADS, jid)
    if not load:
        return {}
    minions = cache.fetch(_BANK_MINIONS, jid)
    if minions:
        load["Minions"] = sorted(minions)
    return load


def get_jid(jid):
    """
    Return ``{minion_id: return_dict}`` for every return recorded
    against *jid*.

    Each return is a dict with at least ``return``; ``retcode``,
    ``success``, and ``out`` are populated when the minion reported
    them (same shape as ``local_cache``).
    """
    cache = _cache()
    returns_bank = _BANK_RETURNS_FMT.format(jid=jid)
    try:
        raw = cache.list_all(returns_bank, include_data=True)
    except (AttributeError, salt.exceptions.SaltCacheError):
        # Fallback for drivers that don't implement list_all (or for
        # a JID that has no returns yet).
        raw = {}
        for minion_id in cache.list(returns_bank):
            record = cache.fetch(returns_bank, minion_id)
            if record is not None:
                raw[minion_id] = record
    ret = {}
    for minion_id, record in (raw or {}).items():
        if not isinstance(record, dict) or "return" not in record:
            # Backwards-compat: ``local_cache`` v1 stored just the
            # return value at this key.  Wrap into the new shape so
            # consumers get a consistent dict.
            record = {"return": record}
        ret[minion_id] = record
    return ret


def get_jids():
    """
    Return ``{jid: jid_info}`` for every load on disk.

    ``jid_info`` is what :func:`salt.utils.jid.format_jid_instance`
    produces — used by the runner / API surfaces (``salt-run
    jobs.list_jobs`` etc.).  When ``job_cache_store_endtime`` is on,
    the matching ``EndTime`` is folded in.
    """
    cache = _cache()
    ret = {}
    for jid in cache.list(_BANK_LOADS):
        load = cache.fetch(_BANK_LOADS, jid)
        if not load:
            continue
        info = salt.utils.jid.format_jid_instance(jid, load)
        if __opts__.get("job_cache_store_endtime"):
            endtime = get_endtime(jid)
            if endtime:
                info["EndTime"] = endtime
        ret[jid] = info
    return ret


def get_jids_filter(count, filter_find_job=True):
    """
    Return the *count* most-recent JIDs as ``jid_info_ext`` dicts,
    optionally filtering out ``saltutil.find_job`` traffic.

    JIDs sort lexicographically by their timestamp prefix so the
    most-recent is the last in sorted order.  We accumulate into a
    bounded list to stay O(N log count) rather than O(N log N).
    """
    cache = _cache()
    keys = []
    ret = []
    import bisect  # pylint: disable=import-outside-toplevel

    for jid in cache.list(_BANK_LOADS):
        load = cache.fetch(_BANK_LOADS, jid)
        if not load:
            continue
        job = salt.utils.jid.format_jid_instance_ext(jid, load)
        if filter_find_job and job.get("Function") == "saltutil.find_job":
            continue
        i = bisect.bisect(keys, jid)
        if len(keys) == count and i == 0:
            continue
        keys.insert(i, jid)
        ret.insert(i, job)
        if len(keys) > count:
            del keys[0]
            del ret[0]
    return ret


# ---------------------------------------------------------------------------
# Endtime helpers — populated only when ``job_cache_store_endtime`` is set
# ---------------------------------------------------------------------------


def update_endtime(jid, the_time):
    """Record (or overwrite) the end-time for *jid*."""
    _cache().store(_BANK_ENDTIMES, jid, the_time)


def get_endtime(jid):
    """Return the recorded end-time for *jid*, or ``None``."""
    return _cache().fetch(_BANK_ENDTIMES, jid) or None


# ---------------------------------------------------------------------------
# Maintenance
# ---------------------------------------------------------------------------


def clean_old_jobs():
    """
    Drop loads (and all their associated banks) older than
    ``keep_jobs_seconds`` opt.

    Uses :meth:`salt.cache.Cache.updated` to age each load; falls
    back to "keep" when the driver can't report a timestamp (some
    drivers don't implement ``updated``).  Mirrors
    ``local_cache.clean_old_jobs`` semantically; the loop is far
    simpler because we don't have to walk a two-level directory
    hierarchy.
    """
    keep_jobs_seconds = salt.utils.job.get_keep_jobs_seconds(__opts__)
    if keep_jobs_seconds == 0:
        return
    cache = _cache()
    now = time.time()
    for jid in cache.list(_BANK_LOADS):
        try:
            updated = cache.updated(_BANK_LOADS, jid)
        except (AttributeError, salt.exceptions.SaltCacheError):
            # Driver can't tell us when this was written — leave the
            # entry alone rather than risk dropping a fresh job.
            continue
        if updated is None:
            continue
        if now - updated <= keep_jobs_seconds:
            continue
        _drop_jid(cache, jid)


def _drop_jid(cache, jid):
    """Remove every per-JID entry across all banks.  Best-effort."""
    cache.flush(_BANK_LOADS, jid)
    cache.flush(_BANK_MINIONS, jid)
    cache.flush(_BANK_ENDTIMES, jid)
    cache.flush(_BANK_NOCACHE, jid)
    # The per-jid returns bank is a whole-bank flush so every minion
    # return for the jid goes in one call.
    try:
        cache.flush(_BANK_RETURNS_FMT.format(jid=jid))
    except salt.exceptions.SaltCacheError:
        # No bank for a jid that never received a return is fine.
        pass



from __future__ import annotations

from virtualenv.activation.via_template import ViaTemplateActivator


class NushellActivator(ViaTemplateActivator):
    def templates(self):
        yield "activate.nu"

    @staticmethod
    def quote(string):
        """
        Nushell supports raw strings like: r###'this is a string'###.

        https://github.com/nushell/nushell.github.io/blob/main/book/working_with_strings.md

        This method finds the maximum continuous sharps in the string and then
        quote it with an extra sharp.
        """
        max_sharps = 0
        current_sharps = 0
        for char in string:
            if char == "#":
                current_sharps += 1
                max_sharps = max(current_sharps, max_sharps)
            else:
                current_sharps = 0
        wrapping = "#" * (max_sharps + 1)
        return f"r{wrapping}'{string}'{wrapping}"

    def replacements(self, creator, dest_folder):  # noqa: ARG002
        return {
            "__VIRTUAL_PROMPT__": "" if self.flag_prompt is None else self.flag_prompt,
            "__VIRTUAL_ENV__": str(creator.dest),
            "__VIRTUAL_NAME__": creator.env_name,
            "__BIN_NAME__": str(creator.bin_dir.relative_to(creator.dest)),
            "__TCL_LIBRARY__": getattr(creator.interpreter, "tcl_lib", None) or "",
            "__TK_LIBRARY__": getattr(creator.interpreter, "tk_lib", None) or "",
        }


__all__ = [
    "NushellActivator",
]



"""
Run child processes as another Windows user.

The main entry points are :func:`runas` (starts the child with
``CreateProcessWithTokenW`` when the minion can obtain a privileged context) and
:func:`runas_unpriv` (uses ``CreateProcessWithLogonW`` as a fallback). Both are
used from :mod:`salt.modules.cmdmod` when ``runas`` is set on Windows.
"""

# Import Python Libraries
import ctypes
import logging
import os
import subprocess
import threading
import time

from salt.exceptions import CommandExecutionError, TimedProcTimeoutError

try:
    import msvcrt

    import pywintypes
    import win32api
    import win32con
    import win32event
    import win32pipe
    import win32process
    import win32profile
    import win32security
    import winerror

    import salt.platform.win

    HAS_WIN32 = True
except ImportError:
    HAS_WIN32 = False


log = logging.getLogger(__name__)


def _preview_cmd(cmd, max_len=500):
    """
    Shorten a command line for safe logging (full line is never executed from
    this helper).

    Args:
        cmd: Command string; non-strings are formatted with ``repr()``.
        max_len: Maximum length before truncation; an ellipsis is appended.

    Returns:
        Truncated string suitable for debug log lines.
    """
    s = cmd if isinstance(cmd, str) else repr(cmd)
    if len(s) > max_len:
        return s[:max_len] + "…"
    return s


# Although utils are often directly imported, it is also possible to use the
# loader.
def __virtual__():
    """
    Load this module only when PyWin32 (and related imports) succeeded.

    Returns:
        The string ``win_runas`` on success, or ``(False, reason)`` when
        required imports failed so the loader should skip the module.
    """
    if not HAS_WIN32:
        return False, "This utility requires pywin32"

    return "win_runas"


def close_handle(handle):
    """
    Tries to close an object handle
    """
    if handle is not None:
        try:
            win32api.CloseHandle(handle)
        except pywintypes.error as exc:
            if exc.winerror != winerror.ERROR_INVALID_HANDLE:
                raise


def split_username(username):
    """
    Parse domain-style account names into a plain user name and domain.

    Accepts UPN (``user@domain``), down-level (``DOMAIN\\user``), or a bare
    name (domain becomes ``.`` for local accounts).

    Args:
        username: Account string; non-strings are coerced with ``str()`` at the
            start of this function.

    Returns:
        A ``(user_name, domain)`` pair suitable for ``LogonUser`` and family.
    """
    domain = "."
    user_name = str(username)
    # Domain users with User Principal Name (UPN): user@DOMAIN
    if "@" in user_name:
        user_name, domain = user_name.split("@", maxsplit=1)
        domain = domain.removesuffix(".local")
    # Domain users with Down-Level Logon Name: DOMAIN\user
    if "\\" in user_name:
        domain, user_name = user_name.split("\\", maxsplit=1)
    return str(user_name), str(domain)


def create_env(user_token, inherit=False, timeout=1):
    """
    Build a user environment block for ``CreateProcess*`` using
    ``CreateEnvironmentBlock``.

    Retries for up to ``timeout`` seconds if the call fails (e.g. logon/logoff
    races). Does not use ``LoadUserProfile`` / ``UnloadUserProfile``: that can
    block for a long time and is not required when the caller already holds a
    user token from ``LogonUser`` or similar.

    Args:
        user_token: Windows access token (handle) for the target user, as used
            by ``win32profile.CreateEnvironmentBlock`` / ``CreateProcess*``.
        inherit: Passed through to ``CreateEnvironmentBlock`` (whether to
            inherit the environment of the process associated with
            ``user_token``; typically ``False`` for an isolated child env).
        timeout: If ``CreateEnvironmentBlock`` raises, retry until this many
            seconds have elapsed since the first attempt, then give up.

    Returns:
        Environment data for the ``environment`` parameter to ``CreateProcess*``,
        or ``None`` if no block was produced and there was no exception to
        re-raise. On repeated failure, the last ``pywintypes.error`` may be
        raised.
    """
    t0 = time.monotonic()
    log.debug("win_runas.create_env: begin inherit=%s timeout=%ss", inherit, timeout)
    start = time.time()
    env = None
    exc = None
    attempt = 0
    while True:
        attempt += 1
        try:
            env = win32profile.CreateEnvironmentBlock(user_token, inherit)
        except pywintypes.error as exc:
            log.debug(
                "win_runas.create_env: CreateEnvironmentBlock attempt %s failed: %s",
                attempt,
                exc,
            )
        else:
            break
        if time.time() - start > timeout:
            log.debug(
                "win_runas.create_env: giving up after %s attempts in %.2fs",
                attempt,
                time.time() - start,
            )
            break
    elapsed = time.monotonic() - t0
    if env is not None:
        log.debug("win_runas.create_env: success in %.3fs", elapsed)
        return env
    log.debug("win_runas.create_env: no environment block after %.3fs", elapsed)
    if exc is not None:
        log.debug("win_runas.create_env: re-raising last CreateEnvironmentBlock error")
        raise exc


def create_default_env(username):
    """
    Creates an environment with default values.
    """
    result = {}

    systemdrive = os.environ.get("SystemDrive", r"C:")
    defaults = {
        "ALLUSERSPROFILE": r"C:\ProgramData",
        "CommonProgramFiles": r"C:\Program Files\Common Files",
        "CommonProgramFiles(x86)": r"C:\Program Files (x86)\Common Files",
        "CommonProgramW6432": r"C:\Program Files\Common Files",
        "ComputerName": None,
        "ComSpec": r"C:\Windows\system32\cmd.exe",
        "DriverData": r"C:\Windows\System32\Drivers\DriverData",
        "NUMBER_OF_PROCESSORS": None,
        "OS": None,
        "Path": r"C:\Windows\System32;C:\Windows;C:\Windows\System32\Wbem;C:\Windows\System32\WindowsPowerShell\v1.0"
        "\\",
        "PATHEXT": r".COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.MSC",
        "PROCESSOR_ARCHITECTURE": None,
        "PROCESSOR_IDENTIFIER": None,
        "PROCESSOR_LEVEL": None,
        "PROCESSOR_REVISION": None,
        "ProgramData": r"C:\ProgramData",
        "ProgramFiles": r"C:\Program Files",
        "ProgramFiles(x86)": r"C:\Program Files (x86)",
        "ProgramW6432": r"C:\Program Files",
        "PROMPT": r"$P$G",
        "PSModulePath": r"%ProgramFiles%\WindowsPowerShell\Modules;%SystemRoot%\system32\WindowsPowerShell\v1.0\Modules",
        "PUBLIC": r"C:\Users\Public",
        "SystemDrive": systemdrive,
        "SystemRoot": r"C:\Windows",
        "USERDOMAIN": None,
        "windir": r"C:\Windows",
    }
    user_specific = {
        "APPDATA": rf"{systemdrive}\Users\{username}\AppData\Roaming",
        "HOMEDRIVE": systemdrive,
        "HOMEPATH": rf"\Users\{username}",
        "LOCALAPPDATA": rf"{systemdrive}\Users\{username}\AppData\Local",
        "TEMP": rf"{systemdrive}\Users\{username}\AppData\Local\Temp",
        "TMP": rf"{systemdrive}\Users\{username}\AppData\Local\Temp",
        "USERNAME": rf"{username}",
        "USERPROFILE": rf"{systemdrive}\Users\{username}",
    }
    # set default variables based on the current user
    for key, val in defaults.items():
        item = os.environ.get(key, val)
        if key is not None:
            result.update({key: item})
    # set user specific variables
    result.update(user_specific)
    return result


def runas(cmd, username, password=None, **kwargs):
    """
    Run ``cmd`` as ``username`` using a logon token and
    ``CreateProcessWithTokenW``.

    The minion must be able to impersonate ``SYSTEM`` (e.g. admin with
    ``SeImpersonatePrivilege``) for the primary path. If that fails, this
    delegates to :func:`runas_unpriv` (logon + ``CreateProcessWithLogonW``) so a
    password is required for non-service accounts in that case.

    The process is created suspended, the primary thread is resumed with
    ``ResumeThread``, and stdout/stderr are read on helper threads while waiting
    so large child output cannot fill the pipe and deadlock the parent.

    Args:
        cmd: Full process command line (string) or argv sequence; lists are
            passed through ``subprocess.list2cmdline`` to build
            ``lpCommandLine``.
        username: Target account; UPN, ``DOMAIN\\user``, or local name. May be
            non-string (e.g. numeric local user name) and is coerced to ``str``.
        password: Password for ``LogonUser`` when required; may be omitted for
            some service accounts and when S4U logon is used.
        cwd: Optional working directory for the child (``lpCurrentDirectory``);
            ``None`` uses the default for the new process.

    Returns:
        Dict with keys: ``pid``, ``retcode`` (if wait succeeded), ``stdout``,
        ``stderr`` (empty strings on read failure).

    Raises:
        salt.exceptions.CommandExecutionError: If the account name cannot be
            resolved (e.g. ``LookupAccountName`` failure).
    """
    # Sometimes this comes in as an int. LookupAccountName can't handle an int
    # Let's make it a string if it's anything other than a string
    if not isinstance(username, str):
        username = str(username)
    # Validate the domain and sid exist for the username
    try:
        _, domain, _ = win32security.LookupAccountName(None, username)
        username, _ = split_username(username)
    except pywintypes.error as exc:
        message = win32api.FormatMessage(exc.winerror).rstrip("\n")
        raise CommandExecutionError(message)

    # Elevate the token from the current process
    access = win32security.TOKEN_QUERY | win32security.TOKEN_ADJUST_PRIVILEGES
    th = win32security.OpenProcessToken(win32api.GetCurrentProcess(), access)
    import salt.platform.win

    salt.platform.win.elevate_token(th)

    # Try to impersonate the SYSTEM user. This process needs to be running as a
    # user who has been granted the SeImpersonatePrivilege, Administrator
    # accounts have this permission by default.
    try:
        impersonation_token = salt.platform.win.impersonate_sid(
            salt.platform.win.SYSTEM_SID,
            session_id=0,
            privs=["SeTcbPrivilege"],
        )
    except OSError:
        log.debug("Unable to impersonate SYSTEM user")
        impersonation_token = None
        win32api.CloseHandle(th)

    if isinstance(cmd, (list, tuple)):
        # CreateProcess parameter lpCommandLine must be a string.
        # Since it is called directly and not via the subprocess module,
        # the arguments must be processed manually.
        cmd = subprocess.list2cmdline(cmd)

    log.debug(
        "win_runas.runas: user=%r domain=%r cwd=%r cmdline=%s",
        username,
        domain,
        kwargs.get("cwd"),
        _preview_cmd(cmd),
    )

    # Impersonation of the SYSTEM user failed. Fallback to an un-privileged
    # runas.
    if not impersonation_token:
        log.debug("No impersonation token, using unprivileged runas")
        return runas_unpriv(cmd, username, password, **kwargs)

    if domain == "NT AUTHORITY":
        # Logon as a system level account, SYSTEM, LOCAL SERVICE, or NETWORK
        # SERVICE.
        user_token = win32security.LogonUser(
            username,
            domain,
            "",
            win32con.LOGON32_LOGON_SERVICE,
            win32con.LOGON32_PROVIDER_DEFAULT,
        )
    elif password:
        # Login with a password.
        user_token = win32security.LogonUser(
            username,
            domain,
            password,
            win32con.LOGON32_LOGON_INTERACTIVE,
            win32con.LOGON32_PROVIDER_DEFAULT,
        )
    else:
        # Login without a password. This always returns an elevated token.
        user_token = salt.platform.win.logon_msv1_s4u(username).Token

    # Get a linked user token to elevate if needed
    elevation_type = win32security.GetTokenInformation(
        user_token, win32security.TokenElevationType
    )
    if elevation_type > 1:
        user_token = win32security.GetTokenInformation(
            user_token, win32security.TokenLinkedToken
        )

    # Elevate the user token
    salt.platform.win.elevate_token(user_token)

    # Make sure the user's token has access to a windows station and desktop
    salt.platform.win.grant_winsta_and_desktop(user_token)

    # Create pipes for standard in, out and error streams
    security_attributes = win32security.SECURITY_ATTRIBUTES()
    security_attributes.bInheritHandle = 1

    stdin_read, stdin_write = win32pipe.CreatePipe(security_attributes, 0)
    stdin_read = salt.platform.win.make_inheritable(stdin_read)

    stdout_read, stdout_write = win32pipe.CreatePipe(security_attributes, 0)
    stdout_write = salt.platform.win.make_inheritable(stdout_write)

    stderr_read, stderr_write = win32pipe.CreatePipe(security_attributes, 0)
    stderr_write = salt.platform.win.make_inheritable(stderr_write)

    # Run the process without showing a window.
    creationflags = (
        win32process.CREATE_NO_WINDOW
        | win32process.CREATE_SUSPENDED
        | win32process.CREATE_UNICODE_ENVIRONMENT
    )

    flags = win32con.STARTF_USESHOWWINDOW | win32con.STARTF_USESTDHANDLES
    startup_args = {
        "dwFlags": flags,
        "hStdInput": stdin_read.handle,
        "hStdOutput": stdout_write.handle,
    }
    if kwargs.get("redirect_stderr", False):
        startup_args.update({"hStdError": stdout_write.handle})
    else:
        startup_args.update({"hStdError": stderr_write.handle})
    startup_info = salt.platform.win.STARTUPINFO(**startup_args)

    # Create the environment for the user
    log.debug("win_runas.runas: building environment block for user token")
    env = create_env(user_token, inherit=False)
    if kwargs.get("env", {}):
        env.update(kwargs["env"])

    # Set an optional timeout
    timeout = kwargs.get("timeout", None)
    if timeout:
        timeout = timeout * 1000
    else:
        timeout = win32event.INFINITE

    wait = not kwargs.get("bg", False)

    hProcess = None
    hThread = None
    result = None
    ret = {}
    try:
        # Start the process in a suspended state.
        log.debug("win_runas.runas: calling CreateProcessWithTokenW (suspended)")
        t0 = time.monotonic()
        process_info = salt.platform.win.CreateProcessWithTokenW(
            int(user_token),
            logonflags=salt.platform.win.LOGON_WITH_PROFILE,
            applicationname=None,
            commandline=cmd,
            creationflags=creationflags,
            environment=env,
            currentdirectory=kwargs.get("cwd"),
            startupinfo=startup_info,
        )
        log.debug(
            "win_runas.runas: CreateProcessWithTokenW returned in %.3fs",
            time.monotonic() - t0,
        )

        hProcess = process_info.hProcess
        hThread = process_info.hThread
        dwProcessId = process_info.dwProcessId
        dwThreadId = process_info.dwThreadId
        log.debug(
            "win_runas.runas: new process pid=%s main_thread_id=%s",
            dwProcessId,
            dwThreadId,
        )

        # We don't use these, so let's close the handle
        salt.platform.win.kernel32.CloseHandle(stdin_write.handle)
        salt.platform.win.kernel32.CloseHandle(stdout_write.handle)
        salt.platform.win.kernel32.CloseHandle(stderr_write.handle)

        ret = {"pid": dwProcessId}

        # CREATE_SUSPENDED: the primary thread must be resumed. ResumeThread on
        # the PROCESS_INFORMATION handle is more reliable than psutil.Process
        # for the suspended thread Windows hands back.
        if hThread:
            resume_count = win32process.ResumeThread(hThread)
            log.debug(
                "win_runas.runas: ResumeThread returned %s (previously-suspended count)",
                resume_count,
            )
            if resume_count in (-1, 0xFFFFFFFF):
                log.error(
                    "win_runas.runas: ResumeThread failed; process may never run; "
                    "if the next line logs an infinite wait, the child is likely still suspended"
                )
            close_handle(hThread)
            hThread = None

        if wait:
            # Drain stdout/stderr on helper threads *before* waiting. If we
            # WaitForSingleObject first then read, the child can fill the 64k
            # pipe and block in WriteFile (e.g. huge ``dir`` output when ``cd``
            # failed) while the parent is stuck in Wait — deadlock.
            stdout_buf = []
            stderr_buf = []

            def _pump_pipecopy(handle, bucket, name):
                """Read one inherited pipe end into *bucket* (list with one bytes)."""
                try:
                    fd = msvcrt.open_osfhandle(int(handle), os.O_RDONLY | os.O_TEXT)
                    with os.fdopen(fd, "rb") as fh:
                        bucket.append(fh.read())
                except OSError as exc:
                    log.debug("win_runas.runas: reading %s pipe: %s", name, exc)
                    bucket.append(b"")

            t_out = threading.Thread(
                target=_pump_pipecopy,
                args=(stdout_read.handle, stdout_buf, "stdout"),
            )
            t_err = threading.Thread(
                target=_pump_pipecopy,
                args=(stderr_read.handle, stderr_buf, "stderr"),
            )
            t_out.start()
            t_err.start()
            log.debug(
                "win_runas.runas: started pipe-drain threads for pid=%s", dwProcessId
            )

            # Wait for the process to exit and get its return code.
            t_wait = time.monotonic()
            result = win32event.WaitForSingleObject(hProcess, timeout)
            if result == win32con.WAIT_TIMEOUT:
                log.warning(
                    "win_runas.runas: pid=%s timed out after %sms, terminating",
                    dwProcessId,
                    timeout,
                )
                win32process.TerminateProcess(hProcess, 1)
            elif result == win32con.WAIT_OBJECT_0:
                log.debug(
                    "win_runas.runas: WaitForSingleObject signalled after %.3fs for pid=%s",
                    time.monotonic() - t_wait,
                    dwProcessId,
                )
                exitcode = win32process.GetExitCodeProcess(hProcess)
                log.debug("win_runas.runas: pid=%s exit code=%s", dwProcessId, exitcode)
                ret["retcode"] = exitcode

            t_out.join()
            t_err.join()
            log.debug(
                "win_runas.runas: pipe-drain threads joined for pid=%s", dwProcessId
            )
            ret["stdout"] = stdout_buf[0] if stdout_buf else b""
            ret["stderr"] = stderr_buf[0] if stderr_buf else b""
    finally:
        close_handle(hProcess)
        close_handle(hThread)
        close_handle(th)
        close_handle(user_token)
        if impersonation_token is not None:
            win32security.RevertToSelf()
        close_handle(impersonation_token)
        close_handle(stdin_read.handle)
        close_handle(stdout_read.handle)
        close_handle(stderr_read.handle)

    if result == win32con.WAIT_TIMEOUT:
        raise TimedProcTimeoutError(
            "{} : Timed out after {} seconds".format(cmd, kwargs["timeout"])
        )

    return ret


def runas_unpriv(cmd, username, password, **kwargs):
    """
    Run ``cmd`` as ``username`` using ``CreateProcessWithLogonW`` (with profile).

    Used when :func:`runas` cannot use ``CreateProcessWithTokenW`` (e.g. no
    SYSTEM impersonation). Requires ``password`` for the target account. Stdin
    is inherited from the minion; stdout/stderr are collected from anonymous
    pipes. Output is read after the process exit wait (acceptable for typical
    output sizes; the privileged :func:`runas` path uses concurrent draining to
    avoid pipe deadlocks on very large streams).

    Args:
        cmd: Command line string, or a list/tuple passed to
            ``subprocess.list2cmdline``.
        username: Target account; coerced to ``str`` if needed.
        password: Password for ``CreateProcessWithLogonW`` (required).
        cwd: Optional working directory for the child; ``None`` for default.

    Returns:
        Dict with ``pid``, ``retcode`` (if the process handle wait succeeded),
        ``stdout``, and ``stderr`` strings.

    Raises:
        salt.exceptions.CommandExecutionError: If the account name cannot be
            resolved.
    """
    # Sometimes this comes in as an int. LookupAccountName can't handle an int
    # Let's make it a string if it's anything other than a string
    if not isinstance(username, str):
        username = str(username)
    # Validate the domain and sid exist for the username
    try:
        _, domain, _ = win32security.LookupAccountName(None, username)
        username, _ = split_username(username)
    except pywintypes.error as exc:
        message = win32api.FormatMessage(exc.winerror).rstrip("\n")
        raise CommandExecutionError(message)

    if isinstance(cmd, (list, tuple)):
        cmd = subprocess.list2cmdline(cmd)
    log.debug(
        "win_runas.runas_unpriv: user=%r domain=%r cwd=%r cmdline=%s",
        username,
        domain,
        kwargs.get("cwd"),
        _preview_cmd(cmd),
    )

    # Create inheritable copy of the stdin
    stdin = salt.platform.win.kernel32.GetStdHandle(
        salt.platform.win.STD_INPUT_HANDLE,
    )
    stdin_read_handle = salt.platform.win.DuplicateHandle(srchandle=stdin, inherit=True)

    # Create a pipe to set as stdout in the child. The write handle needs to be
    # inheritable.
    stdout_read_handle, stdout_write_handle = salt.platform.win.CreatePipe(
        inherit_read=False,
        inherit_write=True,
    )
    stderr_read_handle, stderr_write_handle = salt.platform.win.CreatePipe(
        inherit_read=False,
        inherit_write=True,
    )

    # Run the process without showing a window.
    creationflags = (
        salt.platform.win.CREATE_NO_WINDOW
        | salt.platform.win.CREATE_UNICODE_ENVIRONMENT
    )

    # Get startup info structure
    flags = (
        salt.platform.win.STARTF_USESHOWWINDOW | salt.platform.win.STARTF_USESTDHANDLES
    )
    startup_args = {
        "dwFlags": flags,
        "hStdInput": stdin_read_handle,
        "hStdOutput": stdout_write_handle,
    }
    if kwargs.get("redirect_stderr", False):
        startup_args.update({"hStdError": stdout_write_handle})
    else:
        startup_args.update({"hStdError": stderr_write_handle})
    startup_info = salt.platform.win.STARTUPINFO(**startup_args)

    # Create the environment for the user
    env = kwargs.get("env", None)
    if env:
        # Unprivileged users won't be able to call CreateEnvironmentBlock.
        # Create an environment block with sane defaults instead
        env = create_default_env(username)
        env.update(kwargs["env"])
        env = salt.platform.win.environment_string(env)

    # Set an optional timeout
    timeout = kwargs.get("timeout", None)
    if timeout:
        timeout = timeout * 1000
    else:
        timeout = win32event.INFINITE

    wait = not kwargs.get("bg", False)

    hProcess = None
    hThread = None
    result = None
    ret = {}
    try:
        # Run command and return process info structure
        log.debug("win_runas.runas_unpriv: CreateProcessWithLogonW")
        t0 = time.monotonic()
        process_info = salt.platform.win.CreateProcessWithLogonW(
            username=username,
            domain=domain,
            password=password,
            logonflags=salt.platform.win.LOGON_WITH_PROFILE,
            applicationname=None,
            commandline=cmd,
            creationflags=creationflags,
            environment=env,
            currentdirectory=kwargs.get("cwd"),
            startupinfo=startup_info,
        )
        log.debug(
            "win_runas.runas_unpriv: process started pid=%s in %.3fs",
            process_info.dwProcessId,
            time.monotonic() - t0,
        )
        hProcess = process_info.hProcess
        hThread = process_info.hThread
        dwProcessId = process_info.dwProcessId

        # We don't use these so let's close the handle
        salt.platform.win.kernel32.CloseHandle(stdin_read_handle)
        salt.platform.win.kernel32.CloseHandle(stdout_write_handle)
        salt.platform.win.kernel32.CloseHandle(stderr_write_handle)

        ret = {"pid": dwProcessId}

        if wait:
            # Wait for the process to exit and get its return code
            result = salt.platform.win.kernel32.WaitForSingleObject(hProcess, timeout)
            if result == win32con.WAIT_TIMEOUT:
                log.warning(
                    "win_runas.runas_unpriv: pid=%s timed out after %sms, terminating",
                    dwProcessId,
                    timeout,
                )
                salt.platform.win.kernel32.TerminateProcess(hProcess, 1)
            elif result == win32con.WAIT_OBJECT_0:
                exitcode = salt.platform.win.wintypes.DWORD()
                salt.platform.win.kernel32.GetExitCodeProcess(
                    hProcess, ctypes.byref(exitcode)
                )
                log.debug(
                    "win_runas.runas_unpriv: pid=%s exit code=%s",
                    dwProcessId,
                    exitcode.value,
                )
                ret["retcode"] = exitcode.value

            # Read Standard out
            fd_out = msvcrt.open_osfhandle(stdout_read_handle, os.O_RDONLY | os.O_TEXT)
            with os.fdopen(fd_out, "rb") as f_out:
                stdout = f_out.read()
                ret["stdout"] = stdout

            # Read Standard error
            fd_err = msvcrt.open_osfhandle(stderr_read_handle, os.O_RDONLY | os.O_TEXT)
            with os.fdopen(fd_err, "rb") as f_err:
                stderr = f_err.read()
                ret["stderr"] = stderr
    finally:
        if hProcess is not None:
            salt.platform.win.kernel32.CloseHandle(hProcess)
        if hThread is not None:
            salt.platform.win.kernel32.CloseHandle(hThread)
        if not wait:
            salt.platform.win.kernel32.CloseHandle(stdout_read_handle)
            salt.platform.win.kernel32.CloseHandle(stderr_read_handle)

    if result == win32con.WAIT_TIMEOUT:
        raise TimedProcTimeoutError(
            "{} : Timed out after {} seconds".format(cmd, kwargs["timeout"])
        )

    return ret



"""
The main entry point for salt-api
"""

import asyncio
import logging
import signal

import salt.loader
import salt.utils.process

log = logging.getLogger(__name__)


class RunNetapi(salt.utils.process.SignalHandlingProcess):
    """
    Runner class that's pickable for netapi modules
    """

    def __init__(self, opts, fname, **kwargs):
        super().__init__(**kwargs)
        self.opts = opts
        self.fname = fname

    def run(self):
        netapi = salt.loader.netapi(self.opts)
        netapi_func = netapi[self.fname]
        netapi_func()


class NetapiClient:
    """
    Start each netapi module that is configured to run
    """

    def __init__(self, opts):
        self.opts = opts
        self.process_manager = salt.utils.process.ProcessManager(
            name="NetAPIProcessManager"
        )
        self.netapi = salt.loader.netapi(self.opts)

    def run(self):
        """
        Load and start all available api modules
        """
        if not self.netapi:
            log.error("Did not find any netapi configurations, nothing to start")

        for fun in self.netapi:
            if fun.endswith(".start"):
                name = f"RunNetapi({self.netapi[fun].__module__})"
                log.info("Starting %s", name)
                self.process_manager.add_process(
                    RunNetapi, args=(self.opts, fun), name=name
                )

        # Install the SIGINT/SIGTERM handlers if not done so far
        if signal.getsignal(signal.SIGINT) is signal.SIG_DFL:
            # No custom signal handling was added, install our own
            signal.signal(signal.SIGINT, self._handle_signals)

        if signal.getsignal(signal.SIGTERM) is signal.SIG_DFL:
            # No custom signal handling was added, install our own
            signal.signal(signal.SIGTERM, self._handle_signals)

        asyncio.run(self.process_manager.run())

    def _handle_signals(self, signum, sigframe):
        # escalate the signals to the process manager
        self.process_manager._handle_signals(signum, sigframe)



import struct

from types import ModuleType
from typing import Union

from ._mode_kw import W, W_inverse


class KWPMode(object):
    """Key Wrap with Padding (KWP) mode.

    This is a deterministic Authenticated Encryption (AE) mode
    for protecting cryptographic keys. See `NIST SP800-38F`_.

    It provides both confidentiality and authenticity, and it designed
    so that any bit of the ciphertext depends on all bits of the plaintext.

    This mode is only available for ciphers that operate on 128 bits blocks
    (e.g., AES).

    .. _`NIST SP800-38F`: http://csrc.nist.gov/publications/nistpubs/800-38F/SP-800-38F.pdf

    :undocumented: __init__
    """

    def __init__(self,
                 factory: ModuleType,
                 key: Union[bytes, bytearray]):

        self.block_size = factory.block_size
        if self.block_size != 16:
            raise ValueError("Key Wrap with Padding mode is only available for ciphers"
                             " that operate on 128 bits blocks")

        self._factory = factory
        self._cipher = factory.new(key, factory.MODE_ECB)
        self._done = False

    def seal(self, plaintext: Union[bytes, bytearray]) -> bytes:
        """Encrypt and authenticate (wrap) a cryptographic key.

        Args:
          plaintext:
            The cryptographic key to wrap.

        Returns:
            The wrapped key.
        """

        if self._done:
            raise ValueError("The cipher cannot be used more than once")

        if len(plaintext) == 0:
            raise ValueError("The plaintext must be at least 1 byte")

        if len(plaintext) >= 2 ** 32:
            raise ValueError("The plaintext is too long")

        padlen = (8 - len(plaintext)) % 8
        padded = plaintext + b'\x00' * padlen

        AIV = b'\xA6\x59\x59\xA6' + struct.pack('>I', len(plaintext))

        if len(padded) == 8:
            res = self._cipher.encrypt(AIV + padded)
        else:
            res = W(self._cipher, AIV + padded)

        return res

    def unseal(self, ciphertext: Union[bytes, bytearray]) -> bytes:
        """Decrypt and authenticate (unwrap) a cryptographic key.

        Args:
          ciphertext:
            The cryptographic key to unwrap.
            It must be at least 16 bytes long, and its length
            must be a multiple of 8.

        Returns:
            The original key.

        Raises: ValueError
           If the ciphertext or the key are not valid.
        """

        if self._done:
            raise ValueError("The cipher cannot be used more than once")

        if len(ciphertext) % 8:
            raise ValueError("The ciphertext must have length multiple of 8 bytes")

        if len(ciphertext) < 16:
            raise ValueError("The ciphertext must be at least 24 bytes long")

        if len(ciphertext) == 16:
            S = self._cipher.decrypt(ciphertext)
        else:
            S = W_inverse(self._cipher, ciphertext)

        if S[:4] != b'\xA6\x59\x59\xA6':
            raise ValueError("Incorrect decryption")

        Plen = struct.unpack('>I', S[4:8])[0]

        padlen = len(S) - 8 - Plen
        if padlen < 0 or padlen > 7:
            raise ValueError("Incorrect decryption")

        if S[len(S) - padlen:] != b'\x00' * padlen:
            raise ValueError("Incorrect decryption")

        return S[8:len(S) - padlen]


def _create_kwp_cipher(factory: ModuleType,
                       **kwargs: Union[bytes, bytearray]) -> KWPMode:
    """Create a new block cipher in Key Wrap with Padding mode.

    Args:
      factory:
        A block cipher module, taken from `Cryptodome.Cipher`.
        The cipher must have block length of 16 bytes, such as AES.

    Keywords:
      key:
        The secret key to use to seal or unseal.
    """

    try:
        key = kwargs["key"]
    except KeyError as e:
        raise TypeError("Missing parameter:" + str(e))

    return KWPMode(factory, key)



# This module is part of GitPython and is released under the
# 3-Clause BSD License: https://opensource.org/license/bsd-3-clause/

"""Functions that are supposed to be as fast as possible."""

__all__ = [
    "tree_to_stream",
    "tree_entries_from_data",
    "traverse_trees_recursive",
    "traverse_tree_recursive",
]

from stat import S_ISDIR

from git.compat import safe_decode, defenc

# typing ----------------------------------------------

from typing import (
    Callable,
    List,
    MutableSequence,
    Sequence,
    Tuple,
    TYPE_CHECKING,
    Union,
    overload,
)

if TYPE_CHECKING:
    from _typeshed import ReadableBuffer

    from git import GitCmdObjectDB

EntryTup = Tuple[bytes, int, str]  # Same as TreeCacheTup in tree.py.
EntryTupOrNone = Union[EntryTup, None]

# ---------------------------------------------------


def tree_to_stream(entries: Sequence[EntryTup], write: Callable[["ReadableBuffer"], Union[int, None]]) -> None:
    """Write the given list of entries into a stream using its ``write`` method.

    :param entries:
        **Sorted** list of tuples with (binsha, mode, name).

    :param write:
        A ``write`` method which takes a data string.
    """
    ord_zero = ord("0")
    bit_mask = 7  # 3 bits set.

    for binsha, mode, name in entries:
        mode_str = b""
        for i in range(6):
            mode_str = bytes([((mode >> (i * 3)) & bit_mask) + ord_zero]) + mode_str
        # END for each 8 octal value

        # git slices away the first octal if it's zero.
        if mode_str[0] == ord_zero:
            mode_str = mode_str[1:]
        # END save a byte

        # Here it comes: If the name is actually unicode, the replacement below will not
        # work as the binsha is not part of the ascii unicode encoding - hence we must
        # convert to an UTF-8 string for it to work properly. According to my tests,
        # this is exactly what git does, that is it just takes the input literally,
        # which appears to be UTF-8 on linux.
        if isinstance(name, str):
            name_bytes = name.encode(defenc)
        else:
            name_bytes = name  # type: ignore[unreachable]  # check runtime types - is always str?
        write(b"".join((mode_str, b" ", name_bytes, b"\0", binsha)))
    # END for each item


def tree_entries_from_data(data: bytes) -> List[EntryTup]:
    """Read the binary representation of a tree and returns tuples of
    :class:`~git.objects.tree.Tree` items.

    :param data:
        Data block with tree data (as bytes).

    :return:
        list(tuple(binsha, mode, tree_relative_path), ...)
    """
    ord_zero = ord("0")
    space_ord = ord(" ")
    len_data = len(data)
    i = 0
    out = []
    while i < len_data:
        mode = 0

        # Read Mode
        # Some git versions truncate the leading 0, some don't.
        # The type will be extracted from the mode later.
        while data[i] != space_ord:
            # Move existing mode integer up one level being 3 bits and add the actual
            # ordinal value of the character.
            mode = (mode << 3) + (data[i] - ord_zero)
            i += 1
        # END while reading mode

        # Byte is space now, skip it.
        i += 1

        # Parse name, it is NULL separated.

        ns = i
        while data[i] != 0:
            i += 1
        # END while not reached NULL

        # Default encoding for strings in git is UTF-8.
        # Only use the respective unicode object if the byte stream was encoded.
        name_bytes = data[ns:i]
        name = safe_decode(name_bytes)

        # Byte is NULL, get next 20.
        i += 1
        sha = data[i : i + 20]
        i = i + 20
        out.append((sha, mode, name))
    # END for each byte in data stream
    return out


def _find_by_name(tree_data: MutableSequence[EntryTupOrNone], name: str, is_dir: bool, start_at: int) -> EntryTupOrNone:
    """Return data entry matching the given name and tree mode or ``None``.

    Before the item is returned, the respective data item is set None in the `tree_data`
    list to mark it done.
    """

    try:
        item = tree_data[start_at]
        if item and item[2] == name and S_ISDIR(item[1]) == is_dir:
            tree_data[start_at] = None
            return item
    except IndexError:
        pass
    # END exception handling
    for index, item in enumerate(tree_data):
        if item and item[2] == name and S_ISDIR(item[1]) == is_dir:
            tree_data[index] = None
            return item
        # END if item matches
    # END for each item
    return None


@overload
def _to_full_path(item: None, path_prefix: str) -> None: ...


@overload
def _to_full_path(item: EntryTup, path_prefix: str) -> EntryTup: ...


def _to_full_path(item: EntryTupOrNone, path_prefix: str) -> EntryTupOrNone:
    """Rebuild entry with given path prefix."""
    if not item:
        return item
    return (item[0], item[1], path_prefix + item[2])


def traverse_trees_recursive(
    odb: "GitCmdObjectDB", tree_shas: Sequence[Union[bytes, None]], path_prefix: str
) -> List[Tuple[EntryTupOrNone, ...]]:
    """
    :return:
        List of list with entries according to the given binary tree-shas.

        The result is encoded in a list
        of n tuple|None per blob/commit, (n == len(tree_shas)), where:

        * [0] == 20 byte sha
        * [1] == mode as int
        * [2] == path relative to working tree root

        The entry tuple is ``None`` if the respective blob/commit did not exist in the
        given tree.

    :param tree_shas:
        Iterable of shas pointing to trees. All trees must be on the same level.
        A tree-sha may be ``None``, in which case ``None``.

    :param path_prefix:
        A prefix to be added to the returned paths on this level.
        Set it ``""`` for the first iteration.

    :note:
        The ordering of the returned items will be partially lost.
    """
    trees_data: List[List[EntryTupOrNone]] = []

    nt = len(tree_shas)
    for tree_sha in tree_shas:
        if tree_sha is None:
            data: List[EntryTupOrNone] = []
        else:
            # Make new list for typing as list invariant.
            data = list(tree_entries_from_data(odb.stream(tree_sha).read()))
        # END handle muted trees
        trees_data.append(data)
    # END for each sha to get data for

    out: List[Tuple[EntryTupOrNone, ...]] = []

    # Find all matching entries and recursively process them together if the match is a
    # tree. If the match is a non-tree item, put it into the result.
    # Processed items will be set None.
    for ti, tree_data in enumerate(trees_data):
        for ii, item in enumerate(tree_data):
            if not item:
                continue
            # END skip already done items
            entries: List[EntryTupOrNone]
            entries = [None for _ in range(nt)]
            entries[ti] = item
            _sha, mode, name = item
            is_dir = S_ISDIR(mode)  # Type mode bits

            # Find this item in all other tree data items.
            # Wrap around, but stop one before our current index, hence ti+nt, not
            # ti+1+nt.
            for tio in range(ti + 1, ti + nt):
                tio = tio % nt
                entries[tio] = _find_by_name(trees_data[tio], name, is_dir, ii)

            # END for each other item data
            # If we are a directory, enter recursion.
            if is_dir:
                out.extend(
                    traverse_trees_recursive(
                        odb,
                        [((ei and ei[0]) or None) for ei in entries],
                        path_prefix + name + "/",
                    )
                )
            else:
                out.append(tuple(_to_full_path(e, path_prefix) for e in entries))

            # END handle recursion
            # Finally mark it done.
            tree_data[ii] = None
        # END for each item

        # We are done with one tree, set all its data empty.
        del tree_data[:]
    # END for each tree_data chunk
    return out


def traverse_tree_recursive(odb: "GitCmdObjectDB", tree_sha: bytes, path_prefix: str) -> List[EntryTup]:
    """
    :return:
        List of entries of the tree pointed to by the binary `tree_sha`.

        An entry has the following format:

        * [0] 20 byte sha
        * [1] mode as int
        * [2] path relative to the repository

    :param path_prefix:
        Prefix to prepend to the front of all returned paths.
    """
    entries = []
    data = tree_entries_from_data(odb.stream(tree_sha).read())

    # Unpacking/packing is faster than accessing individual items.
    for sha, mode, name in data:
        if S_ISDIR(mode):
            entries.extend(traverse_tree_recursive(odb, sha, path_prefix + name + "/"))
        else:
            entries.append((sha, mode, path_prefix + name))
    # END for each item

    return entries



"""
Initialization of PostgreSQL data directory
===========================================

The postgres_initdb module is used to initialize the postgresql
data directory.

.. versionadded:: 2016.3.0

.. code-block:: yaml

    pgsql-data-dir:
      postgres_initdb.present:
        - name: /var/lib/pgsql/data
        - auth: password
        - user: postgres
        - password: strong_password
        - encoding: UTF8
        - locale: C
        - runas: postgres
        - checksums: True
        - waldir: /var/postgresql/wal
"""


def __virtual__():
    """
    Only load if the postgres module is present
    """
    if "postgres.datadir_init" not in __salt__:
        return (
            False,
            "Unable to load postgres module.  Make sure `postgres.bins_dir` is set.",
        )
    return True


def present(
    name,
    user=None,
    password=None,
    auth="password",
    encoding="UTF8",
    locale=None,
    runas=None,
    waldir=None,
    checksums=False,
):
    """
    Initialize the PostgreSQL data directory

    name
        The name of the directory to initialize

    user
        The database superuser name

    password
        The password to set for the postgres user

    auth
        The default authentication method for local connections

    encoding
        The default encoding for new databases

    locale
        The default locale for new databases

    waldir
        The transaction log (WAL) directory (default is to keep WAL
        inside the data directory)

        .. versionadded:: 2019.2.0

    checksums
        If True, the cluster will be created with data page checksums.

        .. note:: Data page checksums are supported since PostgreSQL 9.3.

        .. versionadded:: 2019.2.0

    runas
        The system user the operation should be performed on behalf of
    """
    _cmt = f"Postgres data directory {name} is already present"
    ret = {"name": name, "changes": {}, "result": True, "comment": _cmt}

    if not __salt__["postgres.datadir_exists"](name=name):
        if __opts__["test"]:
            ret["result"] = None
            _cmt = f"Postgres data directory {name} is set to be initialized"
            ret["comment"] = _cmt
            return ret

        kwargs = dict(
            user=user,
            password=password,
            auth=auth,
            encoding=encoding,
            locale=locale,
            waldir=waldir,
            checksums=checksums,
            runas=runas,
        )

        if __salt__["postgres.datadir_init"](name, **kwargs):
            _cmt = f"Postgres data directory {name} has been initialized"
            ret["comment"] = _cmt
            ret["changes"][name] = "Present"
        else:
            _cmt = f"Postgres data directory {name} initialization failed"
            ret["result"] = False
            ret["comment"] = _cmt

    return ret



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Callable, Generator, Iterable, Sequence
from typing import Final

from opentelemetry.metrics import (
    CallbackOptions,
    Counter,
    Meter,
    ObservableGauge,
    Observation,
    UpDownCounter,
)

# pylint: disable=invalid-name
CallbackT = (
    Callable[[CallbackOptions], Iterable[Observation]]
    | Generator[Iterable[Observation], CallbackOptions, None]
)


HW_BATTERY_CHARGE: Final = "hw.battery.charge"
"""
Remaining fraction of battery charge
Instrument: gauge
Unit: 1
"""


def create_hw_battery_charge(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Remaining fraction of battery charge"""
    return meter.create_observable_gauge(
        name=HW_BATTERY_CHARGE,
        callbacks=callbacks,
        description="Remaining fraction of battery charge.",
        unit="1",
    )


HW_BATTERY_CHARGE_LIMIT: Final = "hw.battery.charge.limit"
"""
Lower limit of battery charge fraction to ensure proper operation
Instrument: gauge
Unit: 1
"""


def create_hw_battery_charge_limit(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Lower limit of battery charge fraction to ensure proper operation"""
    return meter.create_observable_gauge(
        name=HW_BATTERY_CHARGE_LIMIT,
        callbacks=callbacks,
        description="Lower limit of battery charge fraction to ensure proper operation.",
        unit="1",
    )


HW_BATTERY_TIME_LEFT: Final = "hw.battery.time_left"
"""
Time left before battery is completely charged or discharged
Instrument: gauge
Unit: s
"""


def create_hw_battery_time_left(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Time left before battery is completely charged or discharged"""
    return meter.create_observable_gauge(
        name=HW_BATTERY_TIME_LEFT,
        callbacks=callbacks,
        description="Time left before battery is completely charged or discharged.",
        unit="s",
    )


HW_CPU_SPEED: Final = "hw.cpu.speed"
"""
CPU current frequency
Instrument: gauge
Unit: Hz
"""


def create_hw_cpu_speed(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """CPU current frequency"""
    return meter.create_observable_gauge(
        name=HW_CPU_SPEED,
        callbacks=callbacks,
        description="CPU current frequency.",
        unit="Hz",
    )


HW_CPU_SPEED_LIMIT: Final = "hw.cpu.speed.limit"
"""
CPU maximum frequency
Instrument: gauge
Unit: Hz
"""


def create_hw_cpu_speed_limit(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """CPU maximum frequency"""
    return meter.create_observable_gauge(
        name=HW_CPU_SPEED_LIMIT,
        callbacks=callbacks,
        description="CPU maximum frequency.",
        unit="Hz",
    )


HW_ENERGY: Final = "hw.energy"
"""
Energy consumed by the component
Instrument: counter
Unit: J
"""


def create_hw_energy(meter: Meter) -> Counter:
    """Energy consumed by the component"""
    return meter.create_counter(
        name=HW_ENERGY,
        description="Energy consumed by the component.",
        unit="J",
    )


HW_ERRORS: Final = "hw.errors"
"""
Number of errors encountered by the component
Instrument: counter
Unit: {error}
"""


def create_hw_errors(meter: Meter) -> Counter:
    """Number of errors encountered by the component"""
    return meter.create_counter(
        name=HW_ERRORS,
        description="Number of errors encountered by the component.",
        unit="{error}",
    )


HW_FAN_SPEED: Final = "hw.fan.speed"
"""
Fan speed in revolutions per minute
Instrument: gauge
Unit: rpm
"""


def create_hw_fan_speed(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Fan speed in revolutions per minute"""
    return meter.create_observable_gauge(
        name=HW_FAN_SPEED,
        callbacks=callbacks,
        description="Fan speed in revolutions per minute.",
        unit="rpm",
    )


HW_FAN_SPEED_LIMIT: Final = "hw.fan.speed.limit"
"""
Speed limit in rpm
Instrument: gauge
Unit: rpm
"""


def create_hw_fan_speed_limit(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Speed limit in rpm"""
    return meter.create_observable_gauge(
        name=HW_FAN_SPEED_LIMIT,
        callbacks=callbacks,
        description="Speed limit in rpm.",
        unit="rpm",
    )


HW_FAN_SPEED_RATIO: Final = "hw.fan.speed_ratio"
"""
Fan speed expressed as a fraction of its maximum speed
Instrument: gauge
Unit: 1
"""


def create_hw_fan_speed_ratio(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Fan speed expressed as a fraction of its maximum speed"""
    return meter.create_observable_gauge(
        name=HW_FAN_SPEED_RATIO,
        callbacks=callbacks,
        description="Fan speed expressed as a fraction of its maximum speed.",
        unit="1",
    )


HW_GPU_IO: Final = "hw.gpu.io"
"""
Received and transmitted bytes by the GPU
Instrument: counter
Unit: By
"""


def create_hw_gpu_io(meter: Meter) -> Counter:
    """Received and transmitted bytes by the GPU"""
    return meter.create_counter(
        name=HW_GPU_IO,
        description="Received and transmitted bytes by the GPU.",
        unit="By",
    )


HW_GPU_MEMORY_LIMIT: Final = "hw.gpu.memory.limit"
"""
Size of the GPU memory
Instrument: updowncounter
Unit: By
"""


def create_hw_gpu_memory_limit(meter: Meter) -> UpDownCounter:
    """Size of the GPU memory"""
    return meter.create_up_down_counter(
        name=HW_GPU_MEMORY_LIMIT,
        description="Size of the GPU memory.",
        unit="By",
    )


HW_GPU_MEMORY_USAGE: Final = "hw.gpu.memory.usage"
"""
GPU memory used
Instrument: updowncounter
Unit: By
"""


def create_hw_gpu_memory_usage(meter: Meter) -> UpDownCounter:
    """GPU memory used"""
    return meter.create_up_down_counter(
        name=HW_GPU_MEMORY_USAGE,
        description="GPU memory used.",
        unit="By",
    )


HW_GPU_MEMORY_UTILIZATION: Final = "hw.gpu.memory.utilization"
"""
Fraction of GPU memory used
Instrument: gauge
Unit: 1
"""


def create_hw_gpu_memory_utilization(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Fraction of GPU memory used"""
    return meter.create_observable_gauge(
        name=HW_GPU_MEMORY_UTILIZATION,
        callbacks=callbacks,
        description="Fraction of GPU memory used.",
        unit="1",
    )


HW_GPU_UTILIZATION: Final = "hw.gpu.utilization"
"""
Fraction of time spent in a specific task
Instrument: gauge
Unit: 1
"""


def create_hw_gpu_utilization(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Fraction of time spent in a specific task"""
    return meter.create_observable_gauge(
        name=HW_GPU_UTILIZATION,
        callbacks=callbacks,
        description="Fraction of time spent in a specific task.",
        unit="1",
    )


HW_HOST_AMBIENT_TEMPERATURE: Final = "hw.host.ambient_temperature"
"""
Ambient (external) temperature of the physical host
Instrument: gauge
Unit: Cel
"""


def create_hw_host_ambient_temperature(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Ambient (external) temperature of the physical host"""
    return meter.create_observable_gauge(
        name=HW_HOST_AMBIENT_TEMPERATURE,
        callbacks=callbacks,
        description="Ambient (external) temperature of the physical host.",
        unit="Cel",
    )


HW_HOST_ENERGY: Final = "hw.host.energy"
"""
Total energy consumed by the entire physical host, in joules
Instrument: counter
Unit: J
Note: The overall energy usage of a host MUST be reported using the specific `hw.host.energy` and `hw.host.power` metrics **only**, instead of the generic `hw.energy` and `hw.power` described in the previous section, to prevent summing up overlapping values.
"""


def create_hw_host_energy(meter: Meter) -> Counter:
    """Total energy consumed by the entire physical host, in joules"""
    return meter.create_counter(
        name=HW_HOST_ENERGY,
        description="Total energy consumed by the entire physical host, in joules.",
        unit="J",
    )


HW_HOST_HEATING_MARGIN: Final = "hw.host.heating_margin"
"""
By how many degrees Celsius the temperature of the physical host can be increased, before reaching a warning threshold on one of the internal sensors
Instrument: gauge
Unit: Cel
"""


def create_hw_host_heating_margin(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """By how many degrees Celsius the temperature of the physical host can be increased, before reaching a warning threshold on one of the internal sensors"""
    return meter.create_observable_gauge(
        name=HW_HOST_HEATING_MARGIN,
        callbacks=callbacks,
        description="By how many degrees Celsius the temperature of the physical host can be increased, before reaching a warning threshold on one of the internal sensors.",
        unit="Cel",
    )


HW_HOST_POWER: Final = "hw.host.power"
"""
Instantaneous power consumed by the entire physical host in Watts (`hw.host.energy` is preferred)
Instrument: gauge
Unit: W
Note: The overall energy usage of a host MUST be reported using the specific `hw.host.energy` and `hw.host.power` metrics **only**, instead of the generic `hw.energy` and `hw.power` described in the previous section, to prevent summing up overlapping values.
"""


def create_hw_host_power(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Instantaneous power consumed by the entire physical host in Watts (`hw.host.energy` is preferred)"""
    return meter.create_observable_gauge(
        name=HW_HOST_POWER,
        callbacks=callbacks,
        description="Instantaneous power consumed by the entire physical host in Watts (`hw.host.energy` is preferred).",
        unit="W",
    )


HW_LOGICAL_DISK_LIMIT: Final = "hw.logical_disk.limit"
"""
Size of the logical disk
Instrument: updowncounter
Unit: By
"""


def create_hw_logical_disk_limit(meter: Meter) -> UpDownCounter:
    """Size of the logical disk"""
    return meter.create_up_down_counter(
        name=HW_LOGICAL_DISK_LIMIT,
        description="Size of the logical disk.",
        unit="By",
    )


HW_LOGICAL_DISK_USAGE: Final = "hw.logical_disk.usage"
"""
Logical disk space usage
Instrument: updowncounter
Unit: By
"""


def create_hw_logical_disk_usage(meter: Meter) -> UpDownCounter:
    """Logical disk space usage"""
    return meter.create_up_down_counter(
        name=HW_LOGICAL_DISK_USAGE,
        description="Logical disk space usage.",
        unit="By",
    )


HW_LOGICAL_DISK_UTILIZATION: Final = "hw.logical_disk.utilization"
"""
Logical disk space utilization as a fraction
Instrument: gauge
Unit: 1
"""


def create_hw_logical_disk_utilization(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Logical disk space utilization as a fraction"""
    return meter.create_observable_gauge(
        name=HW_LOGICAL_DISK_UTILIZATION,
        callbacks=callbacks,
        description="Logical disk space utilization as a fraction.",
        unit="1",
    )


HW_MEMORY_SIZE: Final = "hw.memory.size"
"""
Size of the memory module
Instrument: updowncounter
Unit: By
"""


def create_hw_memory_size(meter: Meter) -> UpDownCounter:
    """Size of the memory module"""
    return meter.create_up_down_counter(
        name=HW_MEMORY_SIZE,
        description="Size of the memory module.",
        unit="By",
    )


HW_NETWORK_BANDWIDTH_LIMIT: Final = "hw.network.bandwidth.limit"
"""
Link speed
Instrument: updowncounter
Unit: By/s
"""


def create_hw_network_bandwidth_limit(meter: Meter) -> UpDownCounter:
    """Link speed"""
    return meter.create_up_down_counter(
        name=HW_NETWORK_BANDWIDTH_LIMIT,
        description="Link speed.",
        unit="By/s",
    )


HW_NETWORK_BANDWIDTH_UTILIZATION: Final = "hw.network.bandwidth.utilization"
"""
Utilization of the network bandwidth as a fraction
Instrument: gauge
Unit: 1
"""


def create_hw_network_bandwidth_utilization(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Utilization of the network bandwidth as a fraction"""
    return meter.create_observable_gauge(
        name=HW_NETWORK_BANDWIDTH_UTILIZATION,
        callbacks=callbacks,
        description="Utilization of the network bandwidth as a fraction.",
        unit="1",
    )


HW_NETWORK_IO: Final = "hw.network.io"
"""
Received and transmitted network traffic in bytes
Instrument: counter
Unit: By
"""


def create_hw_network_io(meter: Meter) -> Counter:
    """Received and transmitted network traffic in bytes"""
    return meter.create_counter(
        name=HW_NETWORK_IO,
        description="Received and transmitted network traffic in bytes.",
        unit="By",
    )


HW_NETWORK_PACKETS: Final = "hw.network.packets"
"""
Received and transmitted network traffic in packets (or frames)
Instrument: counter
Unit: {packet}
"""


def create_hw_network_packets(meter: Meter) -> Counter:
    """Received and transmitted network traffic in packets (or frames)"""
    return meter.create_counter(
        name=HW_NETWORK_PACKETS,
        description="Received and transmitted network traffic in packets (or frames).",
        unit="{packet}",
    )


HW_NETWORK_UP: Final = "hw.network.up"
"""
Link status: `1` (up) or `0` (down)
Instrument: updowncounter
Unit: 1
"""


def create_hw_network_up(meter: Meter) -> UpDownCounter:
    """Link status: `1` (up) or `0` (down)"""
    return meter.create_up_down_counter(
        name=HW_NETWORK_UP,
        description="Link status: `1` (up) or `0` (down).",
        unit="1",
    )


HW_PHYSICAL_DISK_ENDURANCE_UTILIZATION: Final = (
    "hw.physical_disk.endurance_utilization"
)
"""
Endurance remaining for this SSD disk
Instrument: gauge
Unit: 1
"""


def create_hw_physical_disk_endurance_utilization(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Endurance remaining for this SSD disk"""
    return meter.create_observable_gauge(
        name=HW_PHYSICAL_DISK_ENDURANCE_UTILIZATION,
        callbacks=callbacks,
        description="Endurance remaining for this SSD disk.",
        unit="1",
    )


HW_PHYSICAL_DISK_SIZE: Final = "hw.physical_disk.size"
"""
Size of the disk
Instrument: updowncounter
Unit: By
"""


def create_hw_physical_disk_size(meter: Meter) -> UpDownCounter:
    """Size of the disk"""
    return meter.create_up_down_counter(
        name=HW_PHYSICAL_DISK_SIZE,
        description="Size of the disk.",
        unit="By",
    )


HW_PHYSICAL_DISK_SMART: Final = "hw.physical_disk.smart"
"""
Value of the corresponding [S.M.A.R.T.](https://wikipedia.org/wiki/S.M.A.R.T.) (Self-Monitoring, Analysis, and Reporting Technology) attribute
Instrument: gauge
Unit: 1
"""


def create_hw_physical_disk_smart(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Value of the corresponding [S.M.A.R.T.](https://wikipedia.org/wiki/S.M.A.R.T.) (Self-Monitoring, Analysis, and Reporting Technology) attribute"""
    return meter.create_observable_gauge(
        name=HW_PHYSICAL_DISK_SMART,
        callbacks=callbacks,
        description="Value of the corresponding [S.M.A.R.T.](https://wikipedia.org/wiki/S.M.A.R.T.) (Self-Monitoring, Analysis, and Reporting Technology) attribute.",
        unit="1",
    )


HW_POWER: Final = "hw.power"
"""
Instantaneous power consumed by the component
Instrument: gauge
Unit: W
Note: It is recommended to report `hw.energy` instead of `hw.power` when possible.
"""


def create_hw_power(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Instantaneous power consumed by the component"""
    return meter.create_observable_gauge(
        name=HW_POWER,
        callbacks=callbacks,
        description="Instantaneous power consumed by the component.",
        unit="W",
    )


HW_POWER_SUPPLY_LIMIT: Final = "hw.power_supply.limit"
"""
Maximum power output of the power supply
Instrument: updowncounter
Unit: W
"""


def create_hw_power_supply_limit(meter: Meter) -> UpDownCounter:
    """Maximum power output of the power supply"""
    return meter.create_up_down_counter(
        name=HW_POWER_SUPPLY_LIMIT,
        description="Maximum power output of the power supply.",
        unit="W",
    )


HW_POWER_SUPPLY_USAGE: Final = "hw.power_supply.usage"
"""
Current power output of the power supply
Instrument: updowncounter
Unit: W
"""


def create_hw_power_supply_usage(meter: Meter) -> UpDownCounter:
    """Current power output of the power supply"""
    return meter.create_up_down_counter(
        name=HW_POWER_SUPPLY_USAGE,
        description="Current power output of the power supply.",
        unit="W",
    )


HW_POWER_SUPPLY_UTILIZATION: Final = "hw.power_supply.utilization"
"""
Utilization of the power supply as a fraction of its maximum output
Instrument: gauge
Unit: 1
"""


def create_hw_power_supply_utilization(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Utilization of the power supply as a fraction of its maximum output"""
    return meter.create_observable_gauge(
        name=HW_POWER_SUPPLY_UTILIZATION,
        callbacks=callbacks,
        description="Utilization of the power supply as a fraction of its maximum output.",
        unit="1",
    )


HW_STATUS: Final = "hw.status"
"""
Operational status: `1` (true) or `0` (false) for each of the possible states
Instrument: updowncounter
Unit: 1
Note: `hw.status` is currently specified as an *UpDownCounter* but would ideally be represented using a [*StateSet* as defined in OpenMetrics](https://github.com/prometheus/OpenMetrics/blob/v1.0.0/specification/OpenMetrics.md#stateset). This semantic convention will be updated once *StateSet* is specified in OpenTelemetry. This planned change is not expected to have any consequence on the way users query their timeseries backend to retrieve the values of `hw.status` over time.
"""


def create_hw_status(meter: Meter) -> UpDownCounter:
    """Operational status: `1` (true) or `0` (false) for each of the possible states"""
    return meter.create_up_down_counter(
        name=HW_STATUS,
        description="Operational status: `1` (true) or `0` (false) for each of the possible states.",
        unit="1",
    )


HW_TAPE_DRIVE_OPERATIONS: Final = "hw.tape_drive.operations"
"""
Operations performed by the tape drive
Instrument: counter
Unit: {operation}
"""


def create_hw_tape_drive_operations(meter: Meter) -> Counter:
    """Operations performed by the tape drive"""
    return meter.create_counter(
        name=HW_TAPE_DRIVE_OPERATIONS,
        description="Operations performed by the tape drive.",
        unit="{operation}",
    )


HW_TEMPERATURE: Final = "hw.temperature"
"""
Temperature in degrees Celsius
Instrument: gauge
Unit: Cel
"""


def create_hw_temperature(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Temperature in degrees Celsius"""
    return meter.create_observable_gauge(
        name=HW_TEMPERATURE,
        callbacks=callbacks,
        description="Temperature in degrees Celsius.",
        unit="Cel",
    )


HW_TEMPERATURE_LIMIT: Final = "hw.temperature.limit"
"""
Temperature limit in degrees Celsius
Instrument: gauge
Unit: Cel
"""


def create_hw_temperature_limit(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Temperature limit in degrees Celsius"""
    return meter.create_observable_gauge(
        name=HW_TEMPERATURE_LIMIT,
        callbacks=callbacks,
        description="Temperature limit in degrees Celsius.",
        unit="Cel",
    )


HW_VOLTAGE: Final = "hw.voltage"
"""
Voltage measured by the sensor
Instrument: gauge
Unit: V
"""


def create_hw_voltage(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Voltage measured by the sensor"""
    return meter.create_observable_gauge(
        name=HW_VOLTAGE,
        callbacks=callbacks,
        description="Voltage measured by the sensor.",
        unit="V",
    )


HW_VOLTAGE_LIMIT: Final = "hw.voltage.limit"
"""
Voltage limit in Volts
Instrument: gauge
Unit: V
"""


def create_hw_voltage_limit(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Voltage limit in Volts"""
    return meter.create_observable_gauge(
        name=HW_VOLTAGE_LIMIT,
        callbacks=callbacks,
        description="Voltage limit in Volts.",
        unit="V",
    )


HW_VOLTAGE_NOMINAL: Final = "hw.voltage.nominal"
"""
Nominal (expected) voltage
Instrument: gauge
Unit: V
"""


def create_hw_voltage_nominal(
    meter: Meter, callbacks: Sequence[CallbackT] | None
) -> ObservableGauge:
    """Nominal (expected) voltage"""
    return meter.create_observable_gauge(
        name=HW_VOLTAGE_NOMINAL,
        callbacks=callbacks,
        description="Nominal (expected) voltage.",
        unit="V",
    )



from __future__ import annotations

import argparse
import sys

from ._implementation import DependencyGroupResolver
from ._toml_compat import tomllib


def main(*, argv: list[str] | None = None) -> None:
    if tomllib is None:
        print(
            "Usage error: dependency-groups CLI requires tomli or Python 3.11+",
            file=sys.stderr,
        )
        raise SystemExit(2)

    parser = argparse.ArgumentParser(
        description=(
            "Lint Dependency Groups for validity. "
            "This will eagerly load and check all of your Dependency Groups."
        )
    )
    parser.add_argument(
        "-f",
        "--pyproject-file",
        default="pyproject.toml",
        help="The pyproject.toml file. Defaults to trying in the current directory.",
    )
    args = parser.parse_args(argv if argv is not None else sys.argv[1:])

    with open(args.pyproject_file, "rb") as fp:
        pyproject = tomllib.load(fp)
    dependency_groups_raw = pyproject.get("dependency-groups", {})

    errors: list[str] = []
    try:
        resolver = DependencyGroupResolver(dependency_groups_raw)
    except (ValueError, TypeError) as e:
        errors.append(f"{type(e).__name__}: {e}")
    else:
        for groupname in resolver.dependency_groups:
            try:
                resolver.resolve(groupname)
            except (LookupError, ValueError, TypeError) as e:
                errors.append(f"{type(e).__name__}: {e}")

    if errors:
        print("errors encountered while examining dependency groups:")
        for msg in errors:
            print(f"  {msg}")
        sys.exit(1)
    else:
        print("ok")
        sys.exit(0)


if __name__ == "__main__":
    main()



"""
A convenience system to manage jobs, both active and already run
"""

import fnmatch
import logging
import os

import salt.client
import salt.minion
import salt.payload
import salt.returners
import salt.utils.args
import salt.utils.files
import salt.utils.jid
import salt.utils.master
from salt.exceptions import SaltClientError

try:
    import dateutil.parser as dateutil_parser

    DATEUTIL_SUPPORT = True
except ImportError:
    DATEUTIL_SUPPORT = False

log = logging.getLogger(__name__)


def master():
    """
    Return the actively executing runners for the master

    CLI Example:

    .. code-block:: bash

        salt-run jobs.master
    """
    return salt.utils.master.get_running_jobs(__opts__)


def active(display_progress=False):
    """
    Return a report on all actively running jobs from a job id centric
    perspective

    CLI Example:

    .. code-block:: bash

        salt-run jobs.active
    """
    ret = {}
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        try:
            active_ = client.cmd("*", "saltutil.running", timeout=__opts__["timeout"])
        except SaltClientError as client_error:
            print(client_error)
            return ret

    if display_progress:
        __jid_event__.fire_event(
            {
                "message": "Attempting to contact minions: {}".format(
                    list(active_.keys())
                )
            },
            "progress",
        )
    for minion, data in active_.items():
        if display_progress:
            __jid_event__.fire_event(
                {"message": f"Received reply from minion {minion}"}, "progress"
            )
        if not isinstance(data, list):
            continue
        for job in data:
            if not job["jid"] in ret:
                ret[job["jid"]] = _format_jid_instance(job["jid"], job)
                ret[job["jid"]].update(
                    {"Running": [{minion: job.get("pid", None)}], "Returned": []}
                )
            else:
                ret[job["jid"]]["Running"].append({minion: job["pid"]})

    mminion = salt.minion.MasterMinion(__opts__)
    for jid in ret:
        returner = _get_returner(
            (__opts__["ext_job_cache"], __opts__["master_job_cache"])
        )
        data = mminion.returners[f"{returner}.get_jid"](jid)
        if data:
            for minion in data:
                if minion not in ret[jid]["Returned"]:
                    ret[jid]["Returned"].append(minion)

    return ret


def lookup_jid(
    jid, ext_source=None, returned=True, missing=False, display_progress=False
):
    """
    Return the printout from a previously executed job

    jid
        The jid to look up.

    ext_source
        The external job cache to use. Default: `None`.

    returned : True
        If ``True``, include the minions that did return from the command.

        .. versionadded:: 2015.8.0

    missing : False
        If ``True``, include the minions that did *not* return from the
        command.

    display_progress : False
        If ``True``, fire progress events.

        .. versionadded:: 2015.5.0

    CLI Example:

    .. code-block:: bash

        salt-run jobs.lookup_jid 20130916125524463507
        salt-run jobs.lookup_jid 20130916125524463507 --out=highstate
    """
    ret = {}
    mminion = salt.minion.MasterMinion(__opts__)
    returner = _get_returner(
        (__opts__["ext_job_cache"], ext_source, __opts__["master_job_cache"])
    )

    try:
        data = list_job(jid, ext_source=ext_source, display_progress=display_progress)
    except TypeError:
        return "Requested returner could not be loaded. No JIDs could be retrieved."

    targeted_minions = data.get("Minions", [])
    returns = data.get("Result", {})

    if returns:
        for minion in returns:
            if display_progress:
                __jid_event__.fire_event({"message": minion}, "progress")
            if "return" in returns[minion]:
                if returned:
                    ret[minion] = returns[minion].get("return")
            else:
                if returned:
                    ret[minion] = returns[minion].get("return")
    if missing:
        for minion_id in (x for x in targeted_minions if x not in returns):
            ret[minion_id] = "Minion did not return"

    # We need to check to see if the 'out' key is present and use it to specify
    # the correct outputter, so we get highstate output for highstate runs.
    try:
        # Check if the return data has an 'out' key. We'll use that as the
        # outputter in the absence of one being passed on the CLI.
        outputter = returns[next(iter(returns))].get("out")
    except (StopIteration, AttributeError):
        outputter = None

    if outputter:
        return {"outputter": outputter, "data": ret}
    else:
        return ret


def list_job(jid, ext_source=None, display_progress=False):
    """
    List a specific job given by its jid

    ext_source
        If provided, specifies which external job cache to use.

    display_progress : False
        If ``True``, fire progress events.

        .. versionadded:: 2015.8.8

    CLI Example:

    .. code-block:: bash

        salt-run jobs.list_job 20130916125524463507
        salt-run jobs.list_job 20130916125524463507 --out=pprint
    """
    ret = {"jid": jid}
    mminion = salt.minion.MasterMinion(__opts__)
    returner = _get_returner(
        (__opts__["ext_job_cache"], ext_source, __opts__["master_job_cache"])
    )
    if display_progress:
        __jid_event__.fire_event(
            {"message": f"Querying returner: {returner}"}, "progress"
        )

    job = mminion.returners[f"{returner}.get_load"](jid)
    ret.update(_format_jid_instance(jid, job))
    ret["Result"] = mminion.returners[f"{returner}.get_jid"](jid)

    fstr = "{}.get_endtime".format(__opts__["master_job_cache"])
    if __opts__.get("job_cache_store_endtime") and fstr in mminion.returners:
        endtime = mminion.returners[fstr](jid)
        if endtime:
            ret["EndTime"] = endtime

    return ret


def list_jobs(
    ext_source=None,
    outputter=None,
    search_metadata=None,
    search_function=None,
    search_target=None,
    start_time=None,
    end_time=None,
    display_progress=False,
):
    """
    List all detectable jobs and associated functions

    ext_source
        If provided, specifies which external job cache to use.

    **FILTER OPTIONS**

    .. note::
        If more than one of the below options are used, only jobs which match
        *all* of the filters will be returned.

    search_metadata
        Specify a dictionary to match to the job's metadata. If any of the
        key-value pairs in this dictionary match, the job will be returned.
        Example:

        .. code-block:: bash

            salt-run jobs.list_jobs search_metadata='{"foo": "bar", "baz": "qux"}'

    search_function
        Can be passed as a string or a list. Returns jobs which match the
        specified function. Globbing is allowed. Example:

        .. code-block:: bash

            salt-run jobs.list_jobs search_function='test.*'
            salt-run jobs.list_jobs search_function='["test.*", "pkg.install"]'

        .. versionchanged:: 2015.8.8
            Multiple targets can now also be passed as a comma-separated list.
            For example:

            .. code-block:: bash

                salt-run jobs.list_jobs search_function='test.*,pkg.install'

    search_target
        Can be passed as a string or a list. Returns jobs which match the
        specified minion name. Globbing is allowed. Example:

        .. code-block:: bash

            salt-run jobs.list_jobs search_target='*.mydomain.tld'
            salt-run jobs.list_jobs search_target='["db*", "myminion"]'

        .. versionchanged:: 2015.8.8
            Multiple targets can now also be passed as a comma-separated list.
            For example:

            .. code-block:: bash

                salt-run jobs.list_jobs search_target='db*,myminion'

    start_time
        Accepts any timestamp supported by the dateutil_ Python module (if this
        module is not installed, this argument will be ignored). Returns jobs
        which started after this timestamp.

    end_time
        Accepts any timestamp supported by the dateutil_ Python module (if this
        module is not installed, this argument will be ignored). Returns jobs
        which started before this timestamp.

    .. _dateutil: https://pypi.python.org/pypi/python-dateutil

    CLI Example:

    .. code-block:: bash

        salt-run jobs.list_jobs
        salt-run jobs.list_jobs search_function='test.*' search_target='localhost' search_metadata='{"bar": "foo"}'
        salt-run jobs.list_jobs start_time='2015, Mar 16 19:00' end_time='2015, Mar 18 22:00'

    """
    returner = _get_returner(
        (__opts__["ext_job_cache"], ext_source, __opts__["master_job_cache"])
    )
    if display_progress:
        __jid_event__.fire_event(
            {"message": f"Querying returner {returner} for jobs."}, "progress"
        )
    mminion = salt.minion.MasterMinion(__opts__)

    ret = mminion.returners[f"{returner}.get_jids"]()

    mret = {}
    for item in ret:
        _match = True
        if search_metadata:
            _match = False
            if "Metadata" in ret[item]:
                if isinstance(search_metadata, dict):
                    for key in search_metadata:
                        if key in ret[item]["Metadata"]:
                            if ret[item]["Metadata"][key] == search_metadata[key]:
                                _match = True
                else:
                    log.info(
                        "The search_metadata parameter must be specified"
                        " as a dictionary.  Ignoring."
                    )
        if search_target and _match:
            _match = False
            if "Target" in ret[item]:
                targets = ret[item]["Target"]
                if isinstance(targets, str):
                    targets = [targets]
                for target in targets:
                    for key in salt.utils.args.split_input(search_target):
                        if fnmatch.fnmatch(target, key):
                            _match = True

        if search_function and _match:
            _match = False
            if "Function" in ret[item]:
                for key in salt.utils.args.split_input(search_function):
                    if fnmatch.fnmatch(ret[item]["Function"], key):
                        _match = True

        if start_time and _match:
            _match = False
            if DATEUTIL_SUPPORT:
                parsed_start_time = dateutil_parser.parse(start_time)
                _start_time = dateutil_parser.parse(ret[item]["StartTime"])
                if _start_time >= parsed_start_time:
                    _match = True
            else:
                log.error(
                    "'dateutil' library not available, skipping start_time comparison."
                )

        if end_time and _match:
            _match = False
            if DATEUTIL_SUPPORT:
                parsed_end_time = dateutil_parser.parse(end_time)
                _start_time = dateutil_parser.parse(ret[item]["StartTime"])
                if _start_time <= parsed_end_time:
                    _match = True
            else:
                log.error(
                    "'dateutil' library not available, skipping end_time comparison."
                )

        if _match:
            mret[item] = ret[item]

    if outputter:
        return {"outputter": outputter, "data": mret}
    else:
        return mret


def list_jobs_filter(
    count, filter_find_job=True, ext_source=None, outputter=None, display_progress=False
):
    """
    List all detectable jobs and associated functions

    ext_source
        The external job cache to use. Default: `None`.

    CLI Example:

    .. code-block:: bash

        salt-run jobs.list_jobs_filter 50
        salt-run jobs.list_jobs_filter 100 filter_find_job=False

    """
    returner = _get_returner(
        (__opts__["ext_job_cache"], ext_source, __opts__["master_job_cache"])
    )
    if display_progress:
        __jid_event__.fire_event(
            {"message": f"Querying returner {returner} for jobs."}, "progress"
        )
    mminion = salt.minion.MasterMinion(__opts__)

    fun = f"{returner}.get_jids_filter"
    if fun not in mminion.returners:
        raise NotImplementedError(f"'{fun}' returner function not implemented yet.")
    ret = mminion.returners[fun](count, filter_find_job)

    if outputter:
        return {"outputter": outputter, "data": ret}
    else:
        return ret


def print_job(jid, ext_source=None):
    """
    Print a specific job's detail given by its jid, including the return data.

    CLI Example:

    .. code-block:: bash

        salt-run jobs.print_job 20130916125524463507
    """
    ret = {}

    returner = _get_returner(
        (__opts__["ext_job_cache"], ext_source, __opts__["master_job_cache"])
    )
    mminion = salt.minion.MasterMinion(__opts__)

    try:
        job = mminion.returners[f"{returner}.get_load"](jid)
        ret[jid] = _format_jid_instance(jid, job)
    except TypeError:
        ret[jid]["Result"] = (
            "Requested returner {} is not available. Jobs cannot be "
            "retrieved. Check master log for details.".format(returner)
        )
        return ret
    ret[jid]["Result"] = mminion.returners[f"{returner}.get_jid"](jid)

    fstr = "{}.get_endtime".format(__opts__["master_job_cache"])
    if __opts__.get("job_cache_store_endtime") and fstr in mminion.returners:
        endtime = mminion.returners[fstr](jid)
        if endtime:
            ret[jid]["EndTime"] = endtime

    return ret


def exit_success(jid, ext_source=None):
    """
    Check if a job has been executed and exit successfully

    jid
        The jid to look up.
    ext_source
        The external job cache to use. Default: `None`.

    CLI Example:

    .. code-block:: bash

        salt-run jobs.exit_success 20160520145827701627
    """
    ret = dict()

    data = list_job(jid, ext_source=ext_source)

    minions = data.get("Minions", [])
    result = data.get("Result", {})

    for minion in minions:
        if minion in result and "return" in result[minion]:
            ret[minion] = True if result[minion]["return"] else False
        else:
            ret[minion] = False

    for minion in result:
        if "return" in result[minion] and result[minion]["return"]:
            ret[minion] = True
    return ret


def last_run(
    ext_source=None,
    outputter=None,
    metadata=None,
    function=None,
    target=None,
    display_progress=False,
):
    """
    .. versionadded:: 2015.8.0

    List all detectable jobs and associated functions

    CLI Example:

    .. code-block:: bash

        salt-run jobs.last_run
        salt-run jobs.last_run target=nodename
        salt-run jobs.last_run function='cmd.run'
        salt-run jobs.last_run metadata="{'foo': 'bar'}"
    """

    if metadata:
        if not isinstance(metadata, dict):
            log.info("The metadata parameter must be specified as a dictionary")
            return False

    _all_jobs = list_jobs(
        ext_source=ext_source,
        outputter=outputter,
        search_metadata=metadata,
        search_function=function,
        search_target=target,
        display_progress=display_progress,
    )
    if _all_jobs:
        last_job = sorted(_all_jobs)[-1]
        return print_job(last_job, ext_source)
    else:
        return False


def _get_returner(returner_types):
    """
    Helper to iterate over returner_types and pick the first one
    """
    for returner in returner_types:
        if returner and returner is not None:
            return returner


def _format_job_instance(job):
    """
    Helper to format a job instance
    """
    if not job:
        ret = {"Error": "Cannot contact returner or no job with this jid"}
        return ret

    ret = {
        "Function": job.get("fun", "unknown-function"),
        "Arguments": list(job.get("arg", [])),
        # unlikely but safeguard from invalid returns
        "Target": job.get("tgt", "unknown-target"),
        "Target-type": job.get("tgt_type", "list"),
        "User": job.get("user", "root"),
    }

    if "metadata" in job:
        ret["Metadata"] = job.get("metadata", {})
    else:
        if "kwargs" in job:
            if "metadata" in job["kwargs"]:
                ret["Metadata"] = job["kwargs"].get("metadata", {})

    if "Minions" in job:
        ret["Minions"] = job["Minions"]
    return ret


def _format_jid_instance(jid, job):
    """
    Helper to format jid instance
    """
    ret = _format_job_instance(job)
    ret.update({"StartTime": salt.utils.jid.jid_to_time(jid)})
    return ret


def _walk_through(job_dir, display_progress=False):
    """
    Walk through the job dir and return jobs
    """
    for top in os.listdir(job_dir):
        t_path = os.path.join(job_dir, top)

        for final in os.listdir(t_path):
            load_path = os.path.join(t_path, final, ".load.p")
            with salt.utils.files.fopen(load_path, "rb") as rfh:
                job = salt.payload.load(rfh)

            if not os.path.isfile(load_path):
                continue

            with salt.utils.files.fopen(load_path, "rb") as rfh:
                job = salt.payload.load(rfh)
            jid = job["jid"]
            if display_progress:
                __jid_event__.fire_event({"message": f"Found JID {jid}"}, "progress")
            yield jid, job, t_path, final



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from collections.abc import Sequence

from opentelemetry.context import Context
from opentelemetry.trace import Link, SpanKind, TraceState, get_current_span
from opentelemetry.util.types import Attributes

from ._composable import ComposableSampler, SamplingIntent
from ._trace_state import OtelTraceState
from ._util import (
    INVALID_THRESHOLD,
    MIN_THRESHOLD,
    is_valid_threshold,
)


class _ComposableParentThreshold(ComposableSampler):
    def __init__(self, root_sampler: ComposableSampler):
        self._root_sampler = root_sampler
        self._description = f"ComposableParentThreshold{{root={root_sampler.get_description()}}}"

    def sampling_intent(
        self,
        parent_ctx: Context | None,
        name: str,
        span_kind: SpanKind | None,
        attributes: Attributes,
        links: Sequence[Link] | None,
        trace_state: TraceState | None = None,
    ) -> SamplingIntent:
        parent_span = get_current_span(parent_ctx)
        parent_span_ctx = parent_span.get_span_context()
        is_root = not parent_span_ctx.is_valid
        if is_root:
            return self._root_sampler.sampling_intent(
                parent_ctx, name, span_kind, attributes, links, trace_state
            )

        ot_trace_state = OtelTraceState.parse(trace_state)

        if is_valid_threshold(ot_trace_state.threshold):
            return SamplingIntent(
                threshold=ot_trace_state.threshold,
                threshold_reliable=True,
            )

        threshold = (
            MIN_THRESHOLD
            if parent_span_ctx.trace_flags.sampled
            else INVALID_THRESHOLD
        )
        return SamplingIntent(threshold=threshold, threshold_reliable=False)

    def get_description(self) -> str:
        return self._description


def composable_parent_threshold(
    root_sampler: ComposableSampler,
) -> ComposableSampler:
    """Returns a consistent sampler that respects the sampling decision of
    the parent span or falls-back to the given sampler if it is a root span.

    - For spans without a parent context, delegate to the root sampler
    - For spans with a parent context, returns a SamplingIntent that propagates the parent's sampling decision
    - Returns the parent's threshold if available; otherwise, if the parent's sampled flag is set,
    returns threshold=0; otherwise, if the parent's sampled flag is not set, no threshold is returned.
    - Sets threshold_reliable to match the parent’s reliability, which is true if the parent had a threshold.
    - Does not add any attributes

    Args:
        root_sampler: The root sampler to use for spans without a parent context.
    """
    return _ComposableParentThreshold(root_sampler)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

import binascii
import itertools
import os

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from tests.utils import load_pkcs1_vectors, load_vectors_from_file


def build_vectors(mgf1alg, hashalg, filename):
    vectors = load_vectors_from_file(filename, load_pkcs1_vectors)

    output = []
    for vector in vectors:
        # RSA keys for this must be long enough to accommodate the length of
        # the underlying hash function. This means we can't use the keys from
        # the sha1 test vectors for sha512 tests because 1024-bit keys are too
        # small. Instead we parse the vectors for the test cases, then
        # generate our own 2048-bit keys for each.
        private, _ = vector
        skey = rsa.generate_private_key(65537, 2048)
        pn = skey.private_numbers()
        examples = private["examples"]
        output.append("# =============================================")
        output.append("# Example")
        output.append("# Public key")
        output.append("# Modulus:")
        output.append(format(pn.public_numbers.n, "x"))
        output.append("# Exponent:")
        output.append(format(pn.public_numbers.e, "x"))
        output.append("# Private key")
        output.append("# Modulus:")
        output.append(format(pn.public_numbers.n, "x"))
        output.append("# Public exponent:")
        output.append(format(pn.public_numbers.e, "x"))
        output.append("# Exponent:")
        output.append(format(pn.d, "x"))
        output.append("# Prime 1:")
        output.append(format(pn.p, "x"))
        output.append("# Prime 2:")
        output.append(format(pn.q, "x"))
        output.append("# Prime exponent 1:")
        output.append(format(pn.dmp1, "x"))
        output.append("# Prime exponent 2:")
        output.append(format(pn.dmq1, "x"))
        output.append("# Coefficient:")
        output.append(format(pn.iqmp, "x"))
        pkey = skey.public_key()
        vectorkey = rsa.RSAPrivateNumbers(
            p=private["p"],
            q=private["q"],
            d=private["private_exponent"],
            dmp1=private["dmp1"],
            dmq1=private["dmq1"],
            iqmp=private["iqmp"],
            public_numbers=rsa.RSAPublicNumbers(
                e=private["public_exponent"], n=private["modulus"]
            ),
        ).private_key()
        count = 1

        for example in examples:
            message = vectorkey.decrypt(
                binascii.unhexlify(example["encryption"]),
                padding.OAEP(
                    mgf=padding.MGF1(algorithm=hashes.SHA1()),
                    algorithm=hashes.SHA1(),
                    label=None,
                ),
            )
            assert message == binascii.unhexlify(example["message"])
            ct = pkey.encrypt(
                message,
                padding.OAEP(
                    mgf=padding.MGF1(algorithm=mgf1alg),
                    algorithm=hashalg,
                    label=None,
                ),
            )
            output.append(
                f"# OAEP Example {count} alg={hashalg.name} "
                f"mgf1={mgf1alg.name}"
            )
            count += 1
            output.append("# Message:")
            output.append(example["message"].decode("utf-8"))
            output.append("# Encryption:")
            output.append(binascii.hexlify(ct).decode("utf-8"))

    return "\n".join(output)


def write_file(data, filename):
    with open(filename, "w") as f:
        f.write(data)


oaep_path = os.path.join(
    "asymmetric", "RSA", "pkcs-1v2-1d2-vec", "oaep-vect.txt"
)
hashalgs = [
    hashes.SHA1(),
    hashes.SHA224(),
    hashes.SHA256(),
    hashes.SHA384(),
    hashes.SHA512(),
]
for hashtuple in itertools.product(hashalgs, hashalgs):
    if isinstance(hashtuple[0], hashes.SHA1) and isinstance(
        hashtuple[1], hashes.SHA1
    ):
        continue

    write_file(
        build_vectors(hashtuple[0], hashtuple[1], oaep_path),
        f"oaep-{hashtuple[0].name}-{hashtuple[1].name}.txt",
    )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
RcodeZero DNS Driver
"""
import re
import json
import hashlib

from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import Provider, RecordType, ZoneDoesNotExistError, ZoneAlreadyExistsError
from libcloud.utils.py3 import httplib
from libcloud.common.base import JsonResponse, ConnectionKey
from libcloud.common.types import InvalidCredsError, MalformedResponseError
from libcloud.common.exceptions import BaseHTTPError

API_HOST = "my.rcodezero.at"

__all__ = [
    "RcodeZeroDNSDriver",
]


class RcodeZeroResponse(JsonResponse):
    def success(self):
        i = int(self.status)

        return 200 <= i <= 299

    def parse_error(self):
        if self.status == httplib.UNAUTHORIZED:
            raise InvalidCredsError("Invalid API key. Check https://my.rcodezero.at/enableapi")

        errors = []
        try:
            body = self.parse_body()
        except MalformedResponseError as e:
            body = "{}: {}".format(e.value, e.body)
        try:
            errors = [body["message"]]
        except TypeError:
            return "%s (HTTP Code: %d)" % (body, self.status)
        except KeyError:
            pass

        return "%s (HTTP Code: %d)" % (" ".join(errors), self.status)


class RcodeZeroConnection(ConnectionKey):
    responseCls = RcodeZeroResponse

    host = API_HOST

    def add_default_headers(self, headers):
        headers["Authorization"] = "Bearer " + self.key
        headers["Accept"] = "application/json"

        return headers


class RcodeZeroDNSDriver(DNSDriver):
    type = Provider.RCODEZERO
    name = "RcodeZero DNS"
    website = "https://www.rcodezero.at/"
    connectionCls = RcodeZeroConnection

    RECORD_TYPE_MAP = {
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.AFSDB: "AFSDB",
        RecordType.ALIAS: "ALIAS",
        RecordType.CERT: "CERT",
        RecordType.CNAME: "CNAME",
        RecordType.DNAME: "DNAME",
        RecordType.DNSKEY: "DNSKEY",
        RecordType.DS: "DS",
        RecordType.HINFO: "HINFO",
        RecordType.KEY: "KEY",
        RecordType.LOC: "LOC",
        RecordType.MX: "MX",
        RecordType.NAPTR: "NAPTR",
        RecordType.NS: "NS",
        RecordType.NSEC: "NSEC",
        RecordType.OPENPGPKEY: "OPENPGPKEY",
        RecordType.PTR: "PTR",
        RecordType.RP: "RP",
        RecordType.RRSIG: "RRSIG",
        RecordType.SOA: "SOA",
        RecordType.SPF: "SPF",
        RecordType.SRV: "SRV",
        RecordType.SSHFP: "SSHFP",
        RecordType.SRV: "SRV",
        RecordType.TLSA: "TLSA",
        RecordType.TXT: "TXT",
    }

    def __init__(
        self,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        api_version="v1",
        **kwargs,
    ):
        """
        :param    key: API token to be used (required)
        :type     key: ``str``

        :param    secret: Password to be used, ignored by RcodeZero
        :type     key: ``str``

        :param    secure: Whether to use HTTPS (default) or HTTP.
        :type     secure: ``bool``

        :param    host: Hostname used for connections.
        :type     host: ``str``

        :param    port: Port used for connections.
        :type     port: ``int``

        :param    api_version: Specifies the API version to use.
                               ``v1`` is currently the only valid
                               option (and default)
        :type     api_version: ``str``

        :return: ``None``
        """

        if api_version == "v1":
            self.api_root = "/api/v1"
        else:
            raise NotImplementedError("Unsupported API version: %s" % api_version)

        super().__init__(key=key, secure=secure, host=host, port=port, **kwargs)

    def create_record(self, name, zone, type, data, extra=None):
        """
        Create a new record in a given, existing zone.

        :param name: name of the new record without the domain name,
                     for example "www".
        :type  name: ``str``

        :param zone: Zone in which the requested record is created.
        :type  zone: :class:`Zone`

        :param type: DNS resource record type (A, AAAA, ...).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depending on the record type).
        :type  data: ``str``

        :param extra: Extra attributes: 'ttl', 'disabled'
        :type   extra: ``dict``

        :rtype: :class:`Record`
        """
        action = "{}/zones/{}/rrsets".format(self.api_root, zone.id)

        payload = self._to_patchrequest(zone.id, None, name, type, data, extra, "add")

        try:
            self.connection.request(action=action, data=json.dumps(payload), method="PATCH")
        except BaseHTTPError as e:
            if e.code == httplib.UNPROCESSABLE_ENTITY and e.message.startswith(
                "Could not find domain"
            ):
                raise ZoneDoesNotExistError(zone_id=zone.id, driver=self, value=e.message)
            raise e

        if extra is not None and extra.get("ttl", None) is not None:
            ttl = extra["ttl"]
        else:
            ttl = None

        return Record(id=None, name=name, data=data, type=type, zone=zone, ttl=ttl, driver=self)

    def create_zone(self, domain, type="master", ttl=None, extra={}):
        """
        Create a new zone.

        :param name: Zone domain name (e.g. example.com)
        :type  name: ``str``

        :param domain: Zone type ('master' / 'slave'). (required).
        :type  domain: :class:`Zone`

        :param ttl: TTL for new records. (optional). Ignored by RcodeZero.
                    RcodeZero uses record specific TTLs.
        :type  ttl: ``int``

        :param extra: Extra attributes: 'masters' (for type=slave):
                    ``extra={'masters': ['193.0.2.2','2001:db8::2']}``
                    sets the Master nameservers for a type=slave zone.
        :type extra: ``dict``

        :rtype: :class:`Zone`
        """
        action = "%s/zones" % (self.api_root)

        if type.lower() == "slave" and (extra is None or extra.get("masters", None) is None):
            msg = "Master IPs required for slave zones"
            raise ValueError(msg)
        payload = {"domain": domain.lower(), "type": type.lower()}
        payload.update(extra)
        zone_id = domain + "."
        try:
            self.connection.request(action=action, data=json.dumps(payload), method="POST")
        except BaseHTTPError as e:
            if e.code == httplib.UNPROCESSABLE_ENTITY and e.message.find(
                "Zone '%s' already configured for your account" % domain
            ):
                raise ZoneAlreadyExistsError(zone_id=zone_id, driver=self, value=e.message)
            raise e

        return Zone(
            id=zone_id,
            domain=domain,
            type=type.lower(),
            ttl=None,
            driver=self,
            extra=extra,
        )

    def update_zone(self, zone, domain, type=None, ttl=None, extra=None):
        """
        Update an existing zone.

        :param zone: Zone to update.
        :type  zone: :class:`Zone`

        :param domain: Zone domain name (e.g. example.com)
        :type  domain: ``str``

        :param type: Zone type ('master' / 'slave').
        :type  type: ``str``

        :param ttl: not supported. RcodeZero uses RRSet-specific TTLs
        :type  ttl: ``int``

        :param extra: Extra attributes: 'masters' (for type=slave)
                     ``extra={'masters': ['193.0.2.2','2001:db8::2']}``
                     sets the Master nameserver addresses for a type=slave zone
        :type extra: ``dict``

        :rtype: :class:`Zone`
        """
        action = "{}/zones/{}".format(self.api_root, domain)

        if type is None:
            type = zone.type

        if type.lower() == "slave" and (extra is None or extra.get("masters", None) is None):
            msg = "Master IPs required for slave zones"
            raise ValueError(msg)
        payload = {"domain": domain.lower(), "type": type.lower()}

        if extra is not None:
            payload.update(extra)
        try:
            self.connection.request(action=action, data=json.dumps(payload), method="PUT")
        except BaseHTTPError as e:
            if e.code == httplib.UNPROCESSABLE_ENTITY and e.message.startswith(
                "Domain '%s' update failed" % domain
            ):
                raise ZoneAlreadyExistsError(zone_id=zone.id, driver=self, value=e.message)
            raise e

        return Zone(id=zone.id, domain=domain, type=type, ttl=None, driver=self, extra=extra)

    def delete_record(self, record):
        """
        Delete a record in a given zone.

        :param record: record to delete (record object)
        :type record: `Record`

        :rtype: ``bool``
        """

        action = "{}/zones/{}/rrsets".format(self.api_root, record.zone.id)

        payload = self._to_patchrequest(
            record.zone.id,
            None,
            record.name,
            record.type,
            record.data,
            record.extra,
            "delete",
        )

        try:
            self.connection.request(action=action, data=json.dumps(payload), method="PATCH")

        except BaseHTTPError as e:
            if e.code == httplib.UNPROCESSABLE_ENTITY and e.message.startswith(
                "Could not find domain"
            ):
                raise ZoneDoesNotExistError(zone_id=record.zone.id, driver=self, value=e.message)
            raise e

        return True

    def delete_zone(self, zone):
        """
        Delete a zone and all its records.

        :param zone: zone to delete
        :type zone: `Zone`

        :rtype: ``bool``
        """
        action = "{}/zones/{}".format(self.api_root, zone.id)
        try:
            self.connection.request(action=action, method="DELETE")
        except BaseHTTPError:
            return False

        return True

    def get_zone(self, zone_id):
        """
        Get a Zone object.

        :param zone_id: name of the zone, for
                        example "example.com".
        :type  zone_id: ``str``

        :rtype: :class:`Zone`
        :raises: ZoneDoesNotExistError: if zone could not be found.
        """
        action = "{}/zones/{}".format(self.api_root, zone_id)
        try:
            response = self.connection.request(action=action, method="GET")
        except BaseHTTPError as e:
            if e.code == httplib.NOT_FOUND:
                raise ZoneDoesNotExistError(zone_id=zone_id, driver=self, value=e.message)
            raise e

        return self._to_zone(response.object)

    def get_record(self, zone_id, record_id):
        """
        Return a Record instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :param record_id: ID of the required record
        :type  record_id: ``str``

        :rtype: :class:`Record`
        """
        records = self.list_records(
            Zone(id=zone_id, domain=zone_id, type=None, ttl=None, driver=self, extra=None)
        )

        foundrecords = list(filter(lambda x: x.id == record_id, records))

        if len(foundrecords) > 0:
            return foundrecords[0]
        else:
            return None

    def list_records(self, zone):
        """
        Return a list of all record objects for the given zone.

        :param zone: Zone object to list records for.
        :type zone: :class:`Zone`

        :return: ``list`` of :class:`Record`
        """
        action = "{}/zones/{}/rrsets?page_size=-1".format(self.api_root, zone.id)
        try:
            response = self.connection.request(action=action, method="GET")
        except BaseHTTPError as e:
            if e.code == httplib.UNPROCESSABLE_ENTITY and e.message.startswith(
                "Could not find domain"
            ):
                raise ZoneDoesNotExistError(zone_id=zone.id, driver=self, value=e.message)
            raise e

        return self._to_records(response.object["data"], zone)

    def list_zones(self):
        """
        Return a list of zone objects for this account.

        :return: ``list`` of :class:`Zone`
        """
        action = "%s/zones?page_size=-1" % (self.api_root)
        response = self.connection.request(action=action, method="GET")

        return self._to_zones(response.object["data"])

    def update_record(self, record, name, type, data, extra=None):
        """
        Update an existing record.

        :param record: Record object to update.
        :type  record: :class:`Record`

        :param name: name of the new record, for example "www".
        :type  name: ``str``

        :param type: DNS resource record type (A, AAAA, ...).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depending on the record type).
        :type  data: ``str``

        :param extra: Extra attributes: 'ttl','disabled' (optional)
        :type   extra: ``dict``

        :rtype: :class:`Record`
        """

        action = "{}/zones/{}/rrsets".format(self.api_root, record.zone.id)

        payload = self._to_patchrequest(
            record.zone.id, record, name, type, data, record.extra, "update"
        )

        try:
            self.connection.request(action=action, data=json.dumps(payload), method="PATCH")

        except BaseHTTPError as e:
            if e.code == httplib.UNPROCESSABLE_ENTITY and e.message.startswith(
                "Could not find domain"
            ):
                raise ZoneDoesNotExistError(zone_id=record.zone.id, driver=self, value=e.message)
            raise e

        if not (extra is None or extra.get("ttl", None) is None):
            ttl = extra["ttl"]
        else:
            ttl = record.ttl

        return Record(
            id=hashlib.md5(str(name + " " + data).encode("utf-8")).hexdigest(),  # nosec
            name=name,
            data=data,
            type=type,
            zone=record.zone,
            driver=self,
            ttl=ttl,
            extra=extra,
        )

    def _to_zone(self, item):
        extra = {}

        for e in [
            "dnssec_status",
            "dnssec_status_detail",
            "dnssec_ksk_status",
            "dnssec_ksk_status_detail",
            "dnssec_ds",
            "dnssec_dnskey",
            "dnssec_safe_to_unsign",
            "dnssec",
            "masters",
            "serial",
            "created",
            "last_check",
        ]:
            if e in item:
                extra[e] = item[e]

        return Zone(
            id=item["domain"],
            domain=item["domain"],
            type=item["type"].lower(),
            ttl=None,
            driver=self,
            extra=extra,
        )

    def _to_zones(self, items):
        zones = []

        for item in items:
            zones.append(self._to_zone(item))

        return zones

    def _to_records(self, items, zone):
        records = []

        for item in items:
            for record in item["records"]:
                extra = {}
                extra["disabled"] = record["disabled"]
                # strip domain and trailing dot from recordname
                recordname = re.sub("." + zone.id + "$", "", item["name"][:-1])
                records.append(
                    Record(
                        id=hashlib.md5(  # nosec
                            str(recordname + " " + record["content"]).encode("utf-8")
                        ).hexdigest(),
                        name=recordname,
                        data=record["content"],
                        type=item["type"],
                        zone=zone,
                        driver=self,
                        ttl=item["ttl"],
                        extra=extra,
                    )
                )

        return records

    # rcodezero supports only rrset, so we must create rrsets from the given
    # record
    def _to_patchrequest(self, zone, record, name, type, data, extra, action):
        rrset = {}

        cur_records = self.list_records(
            Zone(id=zone, domain=None, type=None, ttl=None, driver=self)
        )

        if name != "":
            rrset["name"] = name + "." + zone + "."
        else:
            rrset["name"] = zone + "."

        rrset["type"] = type
        rrset["changetype"] = action
        rrset["records"] = []

        if not (extra is None or extra.get("ttl", None) is None):
            rrset["ttl"] = extra["ttl"]

        content = {}

        if not action == "delete":
            content["content"] = data

            if not (extra is None or extra.get("disabled", None) is None):
                content["disabled"] = extra["disabled"]
            rrset["records"].append(content)
        id = hashlib.md5(str(name + " " + data).encode("utf-8")).hexdigest()  # nosec
        # check if rrset contains more than one record. if yes we need to create an
        # update request

        for r in cur_records:
            if action == "update" and r.id == record.id:
                # do not include records which should be updated in the update
                # request

                continue

            if name == r.name and r.id != id:
                # we have other records with the same name so make an update
                # request
                rrset["changetype"] = "update"
                content = {}
                content["content"] = r.data

                if not (r.extra is None or r.extra.get("disabled", None) is None):
                    content["disabled"] = r.extra["disabled"]
                rrset["records"].append(content)
        request = list()
        request.append(rrset)

        return request
